# ruff: noqa: D100, D103
from __future__ import annotations

import asyncio
import json
from typing import TYPE_CHECKING, Any, cast

from ubo_app.store.core.types import ReportReplayingDoneAction

if TYPE_CHECKING:
    from pathlib import Path

    from ubo_app.store.main import UboStore


async def replay_actions(store: UboStore, path: Path) -> None:
    with path.open('r') as file:
        data = json.load(file)

    for item in data:
        store.dispatch(cast('Any', store.load_object(item)))
        await asyncio.sleep(0.5)
    await asyncio.sleep(1.5)
    store.dispatch(ReportReplayingDoneAction())



"""Utility functions for copying template files to the system and restoring them."""

from __future__ import annotations

import pathlib


def copy_templates(
    templates_path: pathlib.Path,
    *,
    variables: dict[str, str],
) -> None:
    """Copy template files to the system and replace variables."""
    for template_path in templates_path.rglob('*'):
        if template_path.is_file() and template_path.name.endswith('.tmpl'):
            relative_path = template_path.relative_to(templates_path)
            system_path = pathlib.Path('/') / relative_path.with_suffix('')
            backup_path = system_path.with_name(system_path.name + '.bak')
            if system_path.exists() and not backup_path.exists():
                system_path.rename(backup_path)
            with template_path.open('r') as template_file:
                template = template_file.read()
                for key, value in variables.items():
                    template = template.replace(f'{{{{{key}}}}}', value)
                with system_path.open('w') as system_file:
                    system_file.write(template)


def restore_backups(templates_path: pathlib.Path) -> None:
    """Restore backup files created by copy_templates."""
    for template_path in templates_path.rglob('*'):
        if template_path.is_file() and template_path.name.endswith('.tmpl'):
            relative_path = template_path.relative_to(templates_path)
            system_path = pathlib.Path('/') / relative_path.with_suffix('')
            backup_path = system_path.with_name(system_path.name + '.bak')
            if backup_path.exists():
                backup_path.rename(system_path)



# ruff: noqa: D100, D103
from __future__ import annotations

import datetime
import functools
import re

from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.input.types import (
    InputAction,
    InputCancelAction,
    InputDemandAction,
    InputMethod,
    InputProvideAction,
    InputResolveAction,
    InputResult,
    QRCodeInputDescription,
)
from ubo_app.store.main import store
from ubo_app.store.services.camera import (
    CameraAction,
    CameraDetectAction,
    CameraDetectedEvent,
    CameraDetectEvent,
    CameraEvent,
    CameraInstallDriverAction,
    CameraInstallDriverEvent,
    CameraReportBarcodeAction,
    CameraRestoreDefaultAction,
    CameraRestoreDefaultEvent,
    CameraSetAvailableCamerasAction,
    CameraSetIndexAction,
    CameraStartViewfinderAction,
    CameraStartViewfinderEvent,
    CameraState,
    CameraStopViewfinderEvent,
    CameraType,
)
from ubo_app.store.services.keypad import KeypadKeyPressAction
from ubo_app.store.services.notifications import (
    Notification,
    NotificationDispatchItem,
    NotificationDisplayType,
    NotificationsAddAction,
    NotificationsClearByIdAction,
)
from ubo_app.utils.persistent_store import read_from_persistent_store

Action = InitAction | CameraAction | InputAction | KeypadKeyPressAction
DispatchAction = (
    NotificationsAddAction | NotificationsClearByIdAction | InputResolveAction
)


def prompt_notification(description: QRCodeInputDescription) -> NotificationsAddAction:
    return NotificationsAddAction(
        notification=Notification(
            id=f'camera:qrcode:{description.id}',
            icon='󰄀󰐲',
            title=description.title or 'QR Code',
            content=f'[size=18dp]{description.prompt}[/size]',
            display_type=NotificationDisplayType.STICKY,
            is_read=True,
            extra_information=description.instructions,
            expiration_timestamp=datetime.datetime.now(tz=datetime.UTC),
            color='#ffffff',
            actions=[
                NotificationDispatchItem(
                    store_action=CameraStartViewfinderAction(
                        pattern=description.pattern,
                    ),
                    icon='󰄀',
                    close_notification=False,
                ),
            ],
            show_dismiss_action=False,
            dismiss_on_close=True,
            on_close=functools.partial(
                store.dispatch,
                InputCancelAction(id=description.id),
            ),
        ),
    )


def pop_queue(
    state: CameraState,
    *,
    actions: list[DispatchAction] | None = None,
    events: list[CameraEvent] | None = None,
) -> ReducerResult[
    CameraState,
    DispatchAction,
    CameraEvent,
]:
    if len(state.queue) == 0:
        msg = 'Cannot pop from an empty queue in CameraState.'
        raise ValueError(msg)
    actions = actions or []
    events = events or []
    events.append(CameraStopViewfinderEvent())

    actions.append(
        NotificationsClearByIdAction(id=f'camera:qrcode:{state.queue[0].id}'),
    )
    _, *queue = state.queue
    if queue:
        actions.append(prompt_notification(queue[0]))
    return CompleteReducerResult(
        state=state(queue=queue),
        actions=actions,
        events=events,
    )


def reducer(
    state: CameraState | None,
    action: Action,
) -> ReducerResult[
    CameraState,
    DispatchAction,
    CameraEvent,
]:
    if state is None:
        if isinstance(action, InitAction):
            # Load persisted camera index
            selected_camera_index = read_from_persistent_store(
                'camera_selected_index',
                default=0,
                output_type=int,
            )
            return CameraState(
                queue=[],
                selected_camera_index=selected_camera_index,
            )
        raise InitializationActionError(action)

    match action:
        case InputDemandAction(description=QRCodeInputDescription() as description):
            return CompleteReducerResult(
                state=state(queue=[*state.queue, description]),
                actions=[] if state.queue else [prompt_notification(description)],
            )

        case InputResolveAction(id=id):
            if state.queue and state.queue[0].id == id:
                return pop_queue(state)
            return state(
                queue=[
                    description for description in state.queue if description.id != id
                ],
            )

        case CameraInstallDriverAction(make=make, model=model, variant=variant):
            return CompleteReducerResult(
                state=state(camera_type=CameraType(variant)),
                events=[
                    CameraInstallDriverEvent(
                        make=make,
                        model=model,
                        variant=variant,
                    ),
                ],
            )

        case CameraRestoreDefaultAction():
            return CompleteReducerResult(
                state=state(camera_type=CameraType.DEFAULT),
                events=[CameraRestoreDefaultEvent()],
            )

        case CameraStartViewfinderAction(pattern=pattern):
            return CompleteReducerResult(
                state=state,
                events=[CameraStartViewfinderEvent(pattern=pattern)],
            )

        case CameraSetIndexAction(index=index):
            return state(selected_camera_index=index)

        case CameraSetAvailableCamerasAction(available_cameras=available_cameras):
            # If current selection is not in available cameras, select first available
            new_index = state.selected_camera_index
            if (
                available_cameras
                and state.selected_camera_index not in available_cameras
            ):
                new_index = available_cameras[0]
            return state(
                available_cameras=tuple(available_cameras),
                selected_camera_index=new_index,
            )

        case CameraDetectAction():
            return CompleteReducerResult(
                state=state,
                events=[CameraDetectEvent()],
            )

        case CameraDetectedEvent(available_cameras=available_cameras):
            # If current selection is not in available cameras, select first available
            new_index = state.selected_camera_index
            if (
                available_cameras
                and state.selected_camera_index not in available_cameras
            ):
                new_index = available_cameras[0]
            return state(
                available_cameras=tuple(available_cameras),
                selected_camera_index=new_index,
            )

        case CameraReportBarcodeAction(codes=codes) if state.queue:
            if state.queue[0].pattern:
                for code in codes:
                    match_ = re.match(state.queue[0].pattern, code)
                    if match_:
                        return CompleteReducerResult(
                            state=state,
                            actions=[
                                InputProvideAction(
                                    id=state.queue[0].id,
                                    value=code,
                                    result=InputResult(
                                        data={
                                            key.rstrip('_'): value
                                            for key, value in match_.groupdict().items()
                                            if value
                                        },
                                        files={},
                                        method=InputMethod.CAMERA,
                                    ),
                                ),
                            ],
                        )
            else:
                return CompleteReducerResult(
                    state=state,
                    actions=[
                        InputProvideAction(
                            id=state.queue[0].id,
                            value=codes[0] if len(codes) > 0 else '',
                            result=None,
                        ),
                    ],
                )
            return state

        case _:
            return state



"""Utility functions for interacting with the APT package manager."""

from __future__ import annotations

import asyncio
import subprocess

from ubo_app.logger import logger


def install_package(packages: str | list[str], /, *, force: bool = False) -> None:
    """Install a package using the APT package manager."""
    import apt  # pyright: ignore [reportMissingModuleSource]

    package_names = packages if isinstance(packages, list) else [packages]

    cache = apt.Cache()
    logger.info('Installing packages...', extra={'packages': package_names})
    for package_name in package_names:
        if package_name in cache:
            pkg = cache[package_name]
            if not pkg.is_installed or force:
                pkg.mark_install()
                cache.commit()
                logger.info(
                    'Package installed successfully.',
                    extra={'package': package_name},
                )
            else:
                msg = f'Package {package_name} is already installed.'
                raise ValueError(msg)
        else:
            msg = f'Package {package_name} not found in APT cache.'
            raise ValueError(msg)


def uninstall_package(package_name: str) -> None:
    """Uninstall a package using the APT package manager."""
    import apt  # pyright: ignore [reportMissingModuleSource]

    cache = apt.Cache()
    logger.info('Uninstalling package...', extra={'package': package_name})
    if package_name in cache:
        pkg = cache[package_name]
        if pkg.is_installed:
            pkg.mark_delete()
            cache.commit()
            logger.info(
                'Package uninstalled successfully.',
                extra={'package': package_name},
            )
        else:
            msg = f'Package {package_name} is not installed.'
            raise ValueError(msg)
    else:
        msg = f'Package {package_name} not found in APT cache.'
        raise ValueError(msg)


async def is_package_installed(package_name: str) -> bool:
    """Asynchronously check if a package is installed using the APT package manager."""
    try:
        process = await asyncio.create_subprocess_exec(
            '/usr/bin/env',
            'dpkg-query',
            '-W',
            '-f=${Status}',
            package_name,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, _ = await process.communicate()
    except subprocess.SubprocessError:
        return False
    else:
        return b'install ok installed' in stdout



# ruff: noqa: D100
from __future__ import annotations

from typing import TYPE_CHECKING, cast

from picamera2.picamera2 import Picamera2

from ubo_app.logger import logger

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray


class PiCamera2Backend:
    """Camera backend implementation using PiCamera2 for Raspberry Pi."""

    def __init__(self, width: int, height: int, camera_index: int = 0) -> None:
        """Initialize the PiCamera2 backend.

        Args:
            width: Desired frame width
            height: Desired frame height
            camera_index: Camera device index (default: 0)

        """
        self._picamera2: Picamera2 | None = None
        self._width = width
        self._height = height
        self._camera_index = camera_index
        self._initialize()

    def _initialize(self) -> None:
        """Initialize the PiCamera2 instance."""
        try:
            self._picamera2 = Picamera2(self._camera_index)
            preview_config = cast(
                'str',
                self._picamera2.create_still_configuration(
                    {
                        'format': 'RGB888',
                        'size': (self._width, self._height),
                    },
                ),
            )
            self._picamera2.configure(preview_config)
            try:
                self._picamera2.set_controls({'AwbEnable': True})
                from ubo_app.utils.persistent_store import read_from_persistent_store

                camera_type = read_from_persistent_store(
                    'camera_type',
                    default='default',
                )
                if camera_type == 'autofocus':
                    self._picamera2.set_controls({'AfMode': 2, 'AfTrigger': 0})
            except Exception:
                logger.exception('Failed to set camera controls.')
        except IndexError:
            logger.exception('Camera not found.')
            self._picamera2 = None

    def start(self) -> None:
        """Start the camera."""
        if self._picamera2:
            self._picamera2.start()

    def stop(self) -> None:
        """Stop the camera."""
        if self._picamera2:
            self._picamera2.stop()

    def close(self) -> None:
        """Release camera resources."""
        if self._picamera2:
            self._picamera2.close()

    def capture_array(self, stream: str = 'main') -> NDArray[np.uint8] | None:
        """Capture a frame from the camera.

        Args:
            stream: The stream name to capture from

        Returns:
            Numpy array containing the frame data in RGB format

        """
        if self._picamera2:
            return self._picamera2.capture_array(stream)
        return None

    def configure(self, config: dict | str) -> None:
        """Configure the camera.

        Args:
            config: Configuration dictionary or string

        """
        if self._picamera2:
            self._picamera2.configure(config)

    def set_controls(self, controls: dict) -> None:
        """Set camera controls.

        Args:
            controls: Dictionary of control parameters

        """
        if self._picamera2:
            self._picamera2.set_controls(controls)



"""Utility functions for interacting with the EEPROM."""

import json
from functools import cache
from pathlib import Path
from typing import TypedDict

from ubo_app.constants import FORCE_HARDWARE
from ubo_app.logger import logger
from ubo_app.utils import IS_RPI


class SoundTestReport(TypedDict):
    """Sound test report."""

    left: bool
    right: bool


class TemperatureTestReport(TypedDict):
    """Temperature test report."""

    degrees: float


class AmbientTestReport(TypedDict):
    """Ambient test report."""

    works: bool
    baseline: float
    reading: float
    delta: float


class KeypadTestReport(TypedDict):
    """Keypad test report."""

    L1: bool
    L2: bool
    L3: bool
    UP: bool
    DOWN: bool
    BACK: bool
    HOME: bool
    MIC: bool


class LCDTestReport(TypedDict):
    """LCD test report."""

    qrcode: bool
    green: bool
    red: bool
    blue: bool


class LEDTestReport(TypedDict):
    """LED test report."""

    green: bool
    red: bool
    blue: bool


class Device(TypedDict):
    """Device information."""

    model: str
    bus_address: str
    test_result: bool


class EepromDevice(Device):
    """EEPROM device information."""


class SpeakersDevice(Device):
    """Speakers device information."""

    test_report: SoundTestReport


class MicrophonesDevice(Device):
    """Microphones device information."""

    test_report: SoundTestReport


class TemperatureDevice(Device):
    """Temperature device information."""

    test_report: TemperatureTestReport


class AmbientDevice(Device):
    """Ambient device information."""

    test_report: AmbientTestReport


class KeypadDevice(Device):
    """Keypad device information."""

    test_report: KeypadTestReport


class LCDDevice(Device):
    """LCD device information."""

    test_report: LCDTestReport


class LEDDevice(Device):
    """LED device information."""

    test_report: LEDTestReport


class InfraredDevice(Device):
    """Infrared device information."""

    test_result: bool


class I2CBus(TypedDict):
    """I2C bus information."""

    num_devices: int
    scanned_addressed: list[str]
    status: str


class EepromData(TypedDict):
    """EEPROM data."""

    serial_number: str
    eeprom: EepromDevice | None
    speakers: SpeakersDevice | None
    microphones: MicrophonesDevice | None
    temperature: TemperatureDevice | None
    ambient: AmbientDevice | None
    keypad: KeypadDevice | None
    i2c_bus: I2CBus | None
    lcd: LCDDevice | None
    led: LEDDevice | None
    infrared: InfraredDevice | None
    version: str
    timedate: str | None
    test_result: bool | None


UNAVAILABLE_SERIAL_NUMBER = '<not-available>'

EMPTY_EEPROM_DATA: EepromData = {
    'serial_number': UNAVAILABLE_SERIAL_NUMBER,
    'eeprom': None,
    'speakers': None,
    'microphones': None,
    'temperature': None,
    'ambient': None,
    'keypad': None,
    'i2c_bus': None,
    'lcd': None,
    'led': None,
    'infrared': None,
    'version': '0.0.0',
    'timedate': None,
    'test_result': None,
}

FULL_HARDWARE_EEPROM_DATA: EepromData = {
    'serial_number': 'FORCE_HARDWARE',
    'eeprom': {
        'model': '24c32',
        'bus_address': '0x50',
        'test_result': True,
    },
    'speakers': {
        'model': 'wm8960',
        'bus_address': '0x1a',
        'test_result': True,
        'test_report': {'left': True, 'right': True},
    },
    'microphones': {
        'model': 'wm8960',
        'bus_address': '0xa1',
        'test_result': True,
        'test_report': {'left': True, 'right': True},
    },
    'temperature': {
        'model': 'pct2075',
        'bus_address': '0x48',
        'test_result': True,
        'test_report': {'degrees': 26.375},
    },
    'ambient': {
        'model': 'VEML7700',
        'bus_address': '0x10',
        'test_result': True,
        'test_report': {
            'works': True,
            'baseline': 2653.2,
            'reading': 5886.4,
            'delta': 3233.2,
        },
    },
    'keypad': {
        'model': 'aw9523',
        'bus_address': '0x58',
        'test_result': True,
        'test_report': {
            'L1': True,
            'L2': True,
            'L3': True,
            'UP': True,
            'DOWN': True,
            'BACK': True,
            'HOME': True,
            'MIC': True,
        },
    },
    'i2c_bus': {
        'num_devices': 3,
        'scanned_addressed': ['0x10', '0x48', '0x58'],
        'status': 'functional_bus',
    },
    'lcd': {
        'model': 'st7789',
        'bus_address': '',
        'test_result': True,
        'test_report': {
            'qrcode': True,
            'green': True,
            'red': True,
            'blue': True,
        },
    },
    'led': {
        'model': 'neopixel',
        'bus_address': '',
        'test_result': True,
        'test_report': {'green': True, 'red': True, 'blue': True},
    },
    'infrared': {
        'model': 'tsop75238',
        'bus_address': '',
        'test_result': True,
    },
    'version': 'V2',
    'timedate': None,
    'test_result': True,
}


@cache
def get_eeprom_data() -> EepromData:
    """Read the EEPROM data."""
    if FORCE_HARDWARE:
        return FULL_HARDWARE_EEPROM_DATA
    if not IS_RPI:
        return EMPTY_EEPROM_DATA
    try:
        eeprom_json_data = Path(
            '/proc/device-tree/hat/custom_0',
        ).read_text(encoding='utf-8')
        data = json.loads(eeprom_json_data)
    except Exception:
        logger.exception('Failed to read EEPROM data')
        return EMPTY_EEPROM_DATA
    else:
        if 'serial_number' not in data or 'version' not in data:
            logger.debug('Invalid EEPROM data')
            return EMPTY_EEPROM_DATA
        eeprom_data: EepromData = data
        return {**EMPTY_EEPROM_DATA, **eeprom_data}


def read_serial_number() -> str | None:
    """Read the serial number from the EEPROM."""
    data = get_eeprom_data()
    return data.get('serial_number')



# ruff: noqa: D100, D101
from __future__ import annotations

from typing import TYPE_CHECKING

from immutable import Immutable
from redux import BaseAction, BaseEvent

if TYPE_CHECKING:
    from collections.abc import Sequence


class IpAction(BaseAction): ...


class IpEvent(BaseEvent): ...


class IpUpdateInterfacesAction(IpAction):
    interfaces: Sequence[IpNetworkInterface]


class IpSetIsConnectedAction(IpAction):
    is_connected: bool


class IpNetworkInterface(Immutable):
    name: str
    ip_addresses: Sequence[str]


class IpState(Immutable):
    interfaces: Sequence[IpNetworkInterface]
    is_connected: bool | None = None



"""Docker image pull progress calculation utilities."""

from __future__ import annotations

import re
import time
from typing import TypedDict

from ubo_app.logger import logger

# Type alias for service tracker dictionary
ServiceTrackers = dict[str, 'LayerProgressTracker']


class DockerLayerStatus:
    """Docker layer status constants."""

    ALREADY_EXISTS = 'Already exists'
    PULL_COMPLETE = 'Pull complete'
    DOWNLOAD_COMPLETE = 'Download complete'
    DOWNLOADING = 'Downloading'
    EXTRACTING = 'Extracting'
    WAITING = 'Waiting'
    PULLING_FS_LAYER = 'Pulling fs layer'

    # Groupings for convenience
    COMPLETED_STATUSES = (ALREADY_EXISTS, PULL_COMPLETE, DOWNLOAD_COMPLETE)


def parse_byte_size(size_str: str) -> int | None:
    """Parse size string like '23.4MB' to bytes.

    Returns None if parsing fails.
    """
    try:
        size_str = size_str.strip()
        match = re.match(r'([0-9.]+)\s*([KMGT]?B)', size_str, re.IGNORECASE)
        if not match:
            return None

        value = float(match.group(1))
        unit = match.group(2).upper()

        multipliers = {
            'B': 1,
            'KB': 1024,
            'MB': 1024 ** 2,
            'GB': 1024 ** 3,
            'TB': 1024 ** 4,
        }

        return int(value * multipliers.get(unit, 1))
    except (ValueError, AttributeError):
        return None


def parse_progress_string(progress_str: str) -> tuple[int, int] | None:
    """Parse progress string like '23.4MB/100MB' or '[==> ] 23.4MB/100MB'.

    Returns (downloaded_bytes, total_bytes) or None if parsing fails.
    """
    try:
        # Remove progress bar indicators like '[==> ]'
        progress_str = re.sub(r'\[.*?\]\s*', '', progress_str)

        # Match pattern like '23.4MB/100MB'
        match = re.match(
            r'([0-9.]+\s*[KMGT]?B)\s*/\s*([0-9.]+\s*[KMGT]?B)',
            progress_str.strip(),
            re.IGNORECASE,
        )

        if not match:
            return None

        downloaded = parse_byte_size(match.group(1))
        total = parse_byte_size(match.group(2))

        if downloaded is None or total is None or total == 0:
            return None

    except (ValueError, AttributeError):
        return None
    else:
        return (downloaded, total)


class LayerData(TypedDict):
    """Type definition for layer progress data."""

    status: str
    downloaded: int
    total: int


class LayerProgressTracker:
    """Track progress of Docker image pull with hybrid byte/layer-based approach."""

    def __init__(self, image_id: str) -> None:
        """Initialize the tracker."""
        self.image_id = image_id
        self.layers: dict[str, LayerData] = {}
        self.last_update_time = 0.0
        self.update_interval = 1.0  # 1 second debounce

    def update_layer(self, layer_id: str, status: str, progress: str = '') -> None:
        """Update layer status and progress.

        Protects completed layers from regressing to incomplete states.
        """
        if layer_id == 'unknown':
            return

        if layer_id not in self.layers:
            self.layers[layer_id] = {
                'status': status,
                'downloaded': 0,
                'total': 0,
            }
        else:
            # Don't allow completed layers to regress
            current_status = self.layers[layer_id]['status']
            if current_status in DockerLayerStatus.COMPLETED_STATUSES:
                # Layer is already complete, don't update status
                return

        self.layers[layer_id]['status'] = status

        # Try to parse byte progress
        if progress:
            parsed = parse_progress_string(progress)
            if parsed:
                downloaded, total = parsed
                self.layers[layer_id]['downloaded'] = downloaded
                self.layers[layer_id]['total'] = total

    def mark_all_complete(self) -> None:
        """Mark all layers as complete (used when service completes)."""
        for layer_data in self.layers.values():
            layer_data['status'] = DockerLayerStatus.PULL_COMPLETE

    def calculate_layer_based_progress(self) -> float:
        """Calculate progress with layer-count weighting (monotonic).

        Each layer contributes equally (1/N) to total progress.
        Within each layer, progress is calculated from bytes downloaded.

        Example: 5 layers, 2 complete, 3rd at 50% downloaded:
        - Layer 1: 100% → contributes 1.0
        - Layer 2: 100% → contributes 1.0
        - Layer 3: 50% → contributes 0.5
        - Layer 4: 0% → contributes 0.0
        - Layer 5: 0% → contributes 0.0
        - Total: (1.0 + 1.0 + 0.5 + 0.0 + 0.0) / 5 = 50%

        Returns:
            Progress as a value between 0 and 1.

        """
        if not self.layers:
            return 0.0

        total_layer_progress = 0.0

        for layer_data in self.layers.values():
            status = layer_data.get('status', '')

            # Fully complete layers contribute 1.0
            if status in DockerLayerStatus.COMPLETED_STATUSES:
                total_layer_progress += 1.0
            # In-progress layers contribute based on bytes downloaded
            elif status == DockerLayerStatus.DOWNLOADING:
                downloaded = layer_data.get('downloaded', 0)
                total = layer_data.get('total', 0)
                if total > 0:
                    # Byte-based progress for this layer
                    layer_progress = downloaded / total
                    total_layer_progress += layer_progress
                # else: no byte info yet, contributes 0.0
            # Other statuses (Waiting, Pulling fs layer, etc.) contribute 0.0

        total_layers = len(self.layers)
        return total_layer_progress / total_layers if total_layers > 0 else 0.0

    def calculate_progress(self) -> float:
        """Calculate overall progress as a value between 0 and 1.

        Uses byte-based calculation when available, falls back to layer counting.

        DEPRECATED: Use calculate_layer_based_progress() for monotonic progress.
        """
        if not self.layers:
            return 0.0

        # Try byte-based calculation first
        total_bytes = 0
        downloaded_bytes = 0
        layers_with_bytes = 0

        for layer_data in self.layers.values():
            total = layer_data.get('total', 0)
            if isinstance(total, int) and total > 0:
                layers_with_bytes += 1
                total_bytes += total
                downloaded_bytes += layer_data.get('downloaded', 0) or 0

        # If we have byte information for at least some layers, use it
        if layers_with_bytes > 0 and total_bytes > 0:
            return min(downloaded_bytes / total_bytes, 1.0)

        # Fall back to layer-based calculation
        layer_weights = {
            DockerLayerStatus.ALREADY_EXISTS: 1.0,
            DockerLayerStatus.PULL_COMPLETE: 1.0,
            DockerLayerStatus.DOWNLOAD_COMPLETE: 0.9,
            DockerLayerStatus.EXTRACTING: 0.7,
            DockerLayerStatus.DOWNLOADING: 0.3,
            DockerLayerStatus.WAITING: 0.1,
            DockerLayerStatus.PULLING_FS_LAYER: 0.0,
        }

        total_progress = 0.0
        for layer_data in self.layers.values():
            status = layer_data.get('status', '')
            total_progress += layer_weights.get(status, 0.0)

        return min(total_progress / len(self.layers), 1.0) if self.layers else 0.0

    def should_update_ui(self) -> bool:
        """Check if enough time has passed to update UI (debouncing)."""
        current_time = time.time()
        if current_time - self.last_update_time >= self.update_interval:
            self.last_update_time = current_time
            return True
        return False

    def cleanup(self) -> None:
        """Clean up layer data to free memory.

        Call this when the tracker is no longer needed to help garbage collection.
        """
        self.layers.clear()
        self.last_update_time = 0.0


def calculate_layer_progress(layer_data: LayerData) -> float:
    """Calculate progress contribution for a single layer.

    Args:
        layer_data: Dictionary containing layer status and progress info

    Returns:
        Progress value between 0.0 and 1.0 for this layer

    """
    status = layer_data.get('status', '')

    # Fully complete layers contribute 1.0
    if status in DockerLayerStatus.COMPLETED_STATUSES:
        return 1.0

    # In-progress layers contribute based on bytes downloaded
    if status == DockerLayerStatus.DOWNLOADING:
        downloaded = layer_data.get('downloaded', 0)
        total = layer_data.get('total', 0)
        if total > 0:
            return downloaded / total
        return 0.0

    # Extracting layers should count as mostly complete
    if status == DockerLayerStatus.EXTRACTING:
        downloaded = layer_data.get('downloaded', 0)
        total = layer_data.get('total', 0)
        if total > 0:
            return max(downloaded / total, 0.9)
        return 0.9

    # Other statuses contribute 0.0
    return 0.0


def calculate_composition_progress(
    service_trackers: ServiceTrackers,
    completed_services: set[str],
    total_count: int,
) -> float:
    """Calculate overall progress using service-segmented approach.

    Each service owns an equal portion of the progress bar (1/total_count).
    Within each service's portion, progress is calculated based on its announced layers.

    Example with 4 services:
    - Service 1 (database): 0-25% based on its layers
    - Service 2 (immich-server): 25-50% based on its layers
    - Service 3 (redis): 50-75% based on its layers
    - Service 4 (immich-machine-learning): 75-100% based on its layers

    This prevents premature 100% completion when only some services have started.

    Args:
        service_trackers: Dict of LayerProgressTracker for each service
        completed_services: Set of completed service names
        total_count: Total number of services expected

    Returns:
        Overall progress as float between 0.0 and 1.0

    """
    if total_count <= 0:
        return 0.0

    # Each service gets an equal portion of the progress bar
    service_portion = 1.0 / total_count
    total_progress = 0.0

    # Track per-service stats for logging
    service_stats = {}
    for service_name, tracker in service_trackers.items():
        # If service is marked complete, force it to contribute full portion
        if service_name in completed_services:
            service_progress = 1.0
            service_stats[service_name] = {
                'total_layers': len(tracker.layers),
                'service_progress': 1.0,
                'is_complete': True,
                'forced_complete': True,
            }
        else:
            service_layers = tracker.layers.values()
            if not service_layers:
                # Service has no announced layers yet, contributes 0.0
                service_progress = 0.0
            else:
                # Calculate progress within this service based on its layers
                service_layer_progress = 0.0
                completed_count = 0
                downloading_count = 0
                extracting_count = 0

                for layer_data in service_layers:
                    layer_progress = calculate_layer_progress(layer_data)
                    service_layer_progress += layer_progress

                    # Update counters for stats
                    status = layer_data.get('status', '')
                    if status in DockerLayerStatus.COMPLETED_STATUSES:
                        completed_count += 1
                    elif status == DockerLayerStatus.DOWNLOADING:
                        downloading_count += 1
                    elif status == DockerLayerStatus.EXTRACTING:
                        extracting_count += 1

                # Service progress is normalized by its layer count
                total_service_layers = len(service_layers)
                service_progress = (service_layer_progress / total_service_layers
                                  if total_service_layers > 0 else 0.0)

                service_stats[service_name] = {
                    'total_layers': total_service_layers,
                    'completed': completed_count,
                    'downloading': downloading_count,
                    'extracting': extracting_count,
                    'service_progress': service_progress,
                    'is_complete': False,
                }

        # Add this service's contribution (its progress * its portion)
        total_progress += service_progress * service_portion

    logger.debug('Calculating progress', extra={
        'total_services': total_count,
        'tracked_services': len(service_trackers),
        'completed_services': len(completed_services),
        'service_portion': service_portion,
        'service_stats': service_stats,
        'total_progress': total_progress,
    })

    return min(total_progress, 1.0)




"""Wake word recognition mixin abstract base class."""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING, final

from typing_extensions import override

from ubo_app.utils.async_evicting_queue import AsyncEvictingQueue

from .base_class import BaseSpeechRecognitionEngine

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Sequence


class WakeWordRecognitionMixin(BaseSpeechRecognitionEngine, abc.ABC):
    """Mixin for wake word detection functionality."""

    wake_words: Sequence[str] | None = None

    @override
    def __init__(self, *, label: str | None = None) -> None:
        """Initialize wake word recognition mixin."""
        self.woke_word_recognitions_queue: AsyncEvictingQueue[str | None] = (
            AsyncEvictingQueue(maxsize=5)
        )
        super().__init__(label=label)

    def set_wake_words(self, wake_words: Sequence[str] | None) -> None:
        """Set the wake words for detection."""
        self.wake_words = wake_words

        self.decide_running_state()

    @final
    async def wake_word_recogntions(self) -> AsyncGenerator[str, None]:
        """Yield recognized wake words."""
        while wake_word := await self.woke_word_recognitions_queue.get():
            yield wake_word

    @override
    async def report(self, result: str) -> None:
        """Report the recognized speech and check for wake words."""
        for wake_word in self.wake_words or []:
            if wake_word.lower() in result.lower():
                await self.woke_word_recognitions_queue.put(wake_word)
                break
        else:
            await super().report(result)

    @override
    def should_be_running(self) -> bool:
        """Check if the wake word engine should be running."""
        return bool(self.wake_words) or super().should_be_running()



# ruff: noqa: D100, D103
from __future__ import annotations

from dataclasses import replace

from redux import (
    BaseEvent,
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
)

from ubo_app.colors import SUCCESS_COLOR
from ubo_app.store.services.notifications import (
    Importance,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.vscode import (
    VSCodeAction,
    VSCodeDoneDownloadingAction,
    VSCodeLoginEvent,
    VSCodeRestartEvent,
    VSCodeSetPendingAction,
    VSCodeSetStatusAction,
    VSCodeStartDownloadingAction,
    VSCodeState,
)


def reducer(
    state: VSCodeState | None,
    action: VSCodeAction,
) -> (
    CompleteReducerResult[VSCodeState, NotificationsAddAction, BaseEvent] | VSCodeState
):
    if state is None:
        if isinstance(action, InitAction):
            return VSCodeState()
        raise InitializationActionError(action)

    match action:
        case VSCodeStartDownloadingAction():
            return replace(state, is_downloading=True)

        case VSCodeDoneDownloadingAction():
            return replace(state, is_downloading=False)

        case VSCodeSetPendingAction():
            return replace(state, is_pending=True)

        case VSCodeSetStatusAction():
            actions = []
            events = []
            if state.is_logged_in is False and action.is_logged_in:
                actions.append(
                    NotificationsAddAction(
                        notification=Notification(
                            id='vscode:login',
                            title='VSCode',
                            content='Successful Login',
                            icon='󰨞',
                            importance=Importance.MEDIUM,
                            color=SUCCESS_COLOR,
                            display_type=NotificationDisplayType.FLASH,
                        ),
                    ),
                )
                events.extend(
                    [
                        VSCodeLoginEvent(),
                        VSCodeRestartEvent(),
                    ],
                )

            state = replace(
                state,
                is_pending=False,
                is_binary_installed=action.is_binary_installed,
                is_logged_in=action.is_logged_in,
                status=action.status,
            )

            return CompleteReducerResult(state=state, actions=actions, events=events)

        case _:
            return state



"""Imperative input."""

from __future__ import annotations

import asyncio
import datetime
import functools
import inspect
from asyncio import Future
from dataclasses import replace
from typing import TYPE_CHECKING, cast, overload

from typing_extensions import TypeVar

from ubo_app.store.input.types import (
    AvailableInputDescription,
    InputCancelEvent,
    InputDemandAction,
    InputMethod,
    InputProvideEvent,
    InputResult,
)
from ubo_app.store.main import store
from ubo_app.store.services.notifications import (
    Notification,
    NotificationActionItem,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.rgb_ring import RgbRingBlinkAction
from ubo_app.store.services.speech_synthesis import ReadableInformation

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence


ReturnType = TypeVar('ReturnType', infer_variance=True)


METHOD_NAMES = {
    InputMethod.CAMERA: 'camera',
    InputMethod.WEB_DASHBOARD: 'web dashboard',
    InputMethod.PATH_SELECTOR: 'path selector',
}

METHOD_ICONS = {
    InputMethod.CAMERA: '󰄀',
    InputMethod.WEB_DASHBOARD: '󱋆',
    InputMethod.PATH_SELECTOR: '󰈤',
}


async def select_input_description(
    descriptions: Sequence[AvailableInputDescription],
) -> AvailableInputDescription | None:
    """Select the input method."""
    if len(descriptions) == 1:
        return descriptions[0]

    input_methods = [description.input_method for description in descriptions]

    if len(input_methods) == 1:
        return next(
            (
                description
                for description in descriptions
                if description.input_method == input_methods[0]
            ),
            None,
        )

    loop = asyncio.get_running_loop()
    input_method_future: Future[InputMethod] = loop.create_future()

    def set_result(method: InputMethod) -> None:
        loop.call_soon_threadsafe(input_method_future.set_result, method)

    if len(input_methods) == 2:  # noqa: PLR2004
        methods_as_text = f"""either {
            ' or '.join([METHOD_NAMES[method] for method in input_methods])
        }"""
    else:
        methods_as_text = f"""{
            ', '.join([METHOD_NAMES[method] for method in input_methods[:-1]])
        }, or {METHOD_NAMES[input_methods[-1]]}"""

    actions = [
        NotificationActionItem(
            key=METHOD_NAMES[method],
            icon=METHOD_ICONS[method],
            label=METHOD_NAMES[method].capitalize(),
            dismiss_notification=True,
            action=functools.partial(set_result, method),
        )
        for method in InputMethod
        if method in input_methods
    ]

    store.dispatch(
        NotificationsAddAction(
            notification=Notification(
                id='input:method',
                icon='',
                title='Input method',
                content='Do you want to use the camera or the web dashboard?',
                display_type=NotificationDisplayType.STICKY,
                is_read=True,
                extra_information=ReadableInformation(
                    text=f'You can use {methods_as_text} to '
                    'enter this input. Please choose one by pressing one of the '
                    'left buttons.',
                ),
                expiration_timestamp=datetime.datetime.now(tz=datetime.UTC),
                color='#ffffff',
                show_dismiss_action=False,
                dismiss_on_close=True,
                actions=actions,
                on_close=lambda: loop.call_soon_threadsafe(
                    input_method_future.cancel,
                ),
            ),
        ),
    )
    selected_input_method = await input_method_future
    return next(
        (
            description
            for description in descriptions
            if description.input_method == selected_input_method
        ),
        None,
    )


@overload
async def ubo_input(
    *,
    prompt: str | None = None,
    title: str | None = None,
    descriptions: Sequence[AvailableInputDescription],
) -> tuple[str, InputResult]: ...
@overload
async def ubo_input(
    *,
    prompt: str | None = None,
    title: str | None = None,
    descriptions: Sequence[AvailableInputDescription],
    resolver: Callable[[str, InputResult | None], ReturnType]
    | Callable[[str], ReturnType],
) -> ReturnType: ...
async def ubo_input(
    *,
    prompt: str | None = None,
    title: str | None = None,
    descriptions: Sequence[AvailableInputDescription],
    resolver: Callable[[str, InputResult | None], ReturnType]
    | Callable[[str], ReturnType]
    | None = None,
) -> tuple[str, InputResult | None] | ReturnType:
    """Input the user in an imperative way."""
    loop = asyncio.get_running_loop()

    descriptions = [
        replace(
            description,
            title=title if description.title is None else description.title,
            prompt=prompt if description.prompt is None else description.prompt,
        )
        for description in descriptions
    ]

    future: Future[tuple[str, InputResult | None]] = loop.create_future()

    selected_input_description = await select_input_description(descriptions)

    if selected_input_description is None:
        msg = 'No input description selected'
        raise asyncio.CancelledError(msg)

    def handle_input_cancel_event(event: InputCancelEvent) -> None:
        if event.id == selected_input_description.id:
            loop.call_soon_threadsafe(future.cancel)

    def handle_input_provide_event(event: InputProvideEvent) -> None:
        if event.id == selected_input_description.id:
            from kivy.utils import get_color_from_hex

            loop.call_soon_threadsafe(
                future.set_result,
                (event.value, event.result),
            )
            kivy_color = get_color_from_hex('#21E693')
            color = (
                round(kivy_color[0] * 255),
                round(kivy_color[1] * 255),
                round(kivy_color[2] * 255),
            )
            store.dispatch(
                RgbRingBlinkAction(
                    color=color,
                    repetitions=1,
                    wait=200,
                ),
            )

    store.subscribe_event(
        InputProvideEvent,
        handle_input_provide_event,
        keep_ref=False,
    )
    store.subscribe_event(
        InputCancelEvent,
        handle_input_cancel_event,
        keep_ref=False,
    )

    store.dispatch(InputDemandAction(description=selected_input_description))

    result = await future

    if not resolver:
        return result

    if len(inspect.signature(resolver).parameters) == 1:
        return cast('Callable[[str], ReturnType]', resolver)(result[0])

    return cast('Callable[[str, InputResult | None], ReturnType]', resolver)(*result)



"""Side effects for the application."""

from __future__ import annotations

import atexit
import json
import signal
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from redux import FinishAction

from ubo_app.store.core.types import (
    PowerOffEvent,
    RebootEvent,
    ReplayRecordedSequenceEvent,
    ScreenshotEvent,
    SnapshotEvent,
    StoreRecordedSequenceEvent,
)
from ubo_app.store.main import store
from ubo_app.store.services.audio import AudioPlayChimeAction
from ubo_app.store.services.notifications import Chime
from ubo_app.store.update_manager.types import (
    UpdateManagerCheckEvent,
    UpdateManagerRequestCheckAction,
    UpdateManagerUpdateEvent,
)
from ubo_app.store.update_manager.utils import check_version, update
from ubo_app.utils import bus_provider
from ubo_app.utils.hardware import IS_RPI, deinitalize_board, initialize_board
from ubo_app.utils.persistent_store import register_persistent_store
from ubo_app.utils.store import replay_actions

if TYPE_CHECKING:
    from numpy._typing._array_like import NDArray

    from ubo_app.utils.types import Subscriptions


def _power_off() -> None:
    """Power off the device."""
    store.dispatch(AudioPlayChimeAction(name=Chime.FAILURE), FinishAction())
    if IS_RPI:

        def power_off_system(*_: list[object]) -> None:
            atexit.unregister(power_off_system)
            atexit._run_exitfuncs()  # noqa: SLF001
            subprocess.run(
                ['/usr/bin/env', 'systemctl', 'poweroff', '-i'],
                check=True,
            )

        atexit.register(power_off_system)


def _reboot() -> None:
    """Reboot the device."""
    store.dispatch(AudioPlayChimeAction(name=Chime.FAILURE), FinishAction())
    if IS_RPI:

        def reboot_system(*_: list[object]) -> None:
            atexit.unregister(reboot_system)
            atexit._run_exitfuncs()  # noqa: SLF001
            subprocess.run(
                ['/usr/bin/env', 'systemctl', 'reboot', '-i'],
                check=True,
            )

        atexit.register(reboot_system)


def _write_image(image_path: Path, array: NDArray) -> None:
    """Write the `NDAarray` as an image to the given path."""
    import png

    array = np.flipud(array)

    png.Writer(
        alpha=True,
        width=array.shape[0],
        height=array.shape[1],
        greyscale=False,  # pyright: ignore [reportArgumentType]
        bitdepth=8,
    ).write(
        image_path.open('wb'),
        array.reshape(-1, array.shape[1] * 4).tolist(),
    )


def _take_screenshot() -> None:
    """Take a screenshot of the screen."""
    from headless_kivy import HeadlessWidget

    counter = 0
    while (path := Path(f'screenshots/ubo-screenshot-{counter:03d}.png')).exists():
        counter += 1

    path.parent.mkdir(parents=True, exist_ok=True)
    _write_image(path, HeadlessWidget.raw_data)


def _take_snapshot() -> None:
    """Take a snapshot of the store."""
    counter = 0
    while (path := Path(f'snapshots/ubo-screenshot-{counter:03d}.json')).exists():
        counter += 1

    path.parent.mkdir(parents=True, exist_ok=True)
    with path.open('w') as file:
        json.dump(store.snapshot, file, indent=2)


def _store_recorded_sequence(event: StoreRecordedSequenceEvent) -> None:
    """Store the recorded sequence."""
    counter = 0
    while (path := Path(f'recordings/ubo-recording-{counter:03d}.json')).exists():
        counter += 1

    path.parent.mkdir(parents=True, exist_ok=True)
    json_dump = json.dumps(
        [
            store.serialize_value(action)
            for action in event.recorded_sequence
            if type(action).__name__.startswith('Keypad')
        ],
        indent=2,
    )

    with path.open('w') as file:
        file.write(json_dump)
    with Path('recordings/active.json').open('w') as file:
        file.write(json_dump)


async def _replay_recorded_sequence() -> None:
    """Replay the recorded sequence."""
    await replay_actions(store, Path('recordings/active.json'))


def setup_side_effects() -> Subscriptions:
    """Set up the side effects for the application."""
    initialize_board()

    register_persistent_store(
        'services',
        lambda state: None
        if state.settings.services is None
        else [
            {
                'id': service.id,
                'is_enabled': service.is_enabled,
                'log_level': service.log_level,
                'should_auto_restart': service.should_auto_restart,
            }
            for service in state.settings.services.values()
        ],
    )
    register_persistent_store(
        'settings:pdb_signal',
        lambda state: state.settings.pdb_signal,
    )
    register_persistent_store(
        'settings:visual_debug',
        lambda state: state.settings.visual_debug,
    )
    register_persistent_store(
        'settings:beta_versions',
        lambda state: state.settings.beta_versions,
    )
    subscriptions = [
        store.subscribe_event(PowerOffEvent, _power_off),
        store.subscribe_event(RebootEvent, _reboot),
        store.subscribe_event(UpdateManagerUpdateEvent, update),
        store.subscribe_event(UpdateManagerCheckEvent, check_version),
        store.subscribe_event(ScreenshotEvent, _take_screenshot),
        store.subscribe_event(SnapshotEvent, _take_snapshot),
        store.subscribe_event(StoreRecordedSequenceEvent, _store_recorded_sequence),
        store.subscribe_event(ReplayRecordedSequenceEvent, _replay_recorded_sequence),
        deinitalize_board,
        bus_provider.clean_up,
    ]

    from kivy.clock import mainthread

    @store.autorun(lambda state: state.settings.pdb_signal)
    @mainthread
    def _pdb_debug_mode(pdb_signal: bool) -> None:  # noqa: FBT001
        """Set the PDB debug mode."""

        def signal_handler(signum: int, _: object) -> None:
            if signum == signal.SIGUSR1:
                import ipdb  # noqa: T100

                ipdb.set_trace()  # noqa: T100
                return

        if pdb_signal:
            signal.signal(signal.SIGUSR1, signal_handler)
        else:
            signal.signal(signal.SIGUSR1, signal.SIG_DFL)

    store.dispatch(UpdateManagerRequestCheckAction())

    return subscriptions



"""Infrared receiver for system manager."""

import asyncio
import functools
import queue
import re
from collections.abc import Iterator
from pathlib import Path
from threading import Thread

from ubo_app.logger import get_logger

logger = get_logger('system-manager')


class InfraredManager:
    """Manager for infrared functionality."""

    def __init__(self) -> None:
        """Initialize the infrared manager."""
        self.loop = asyncio.new_event_loop()
        self.event_loop_thread: Thread | None = None
        self.ir_code_queue = queue.Queue(1)
        self.stop_event = asyncio.Event()

    def handle_command(self, command: str) -> Iterator[str] | str | None:
        """Handle infrared commands."""
        logger.info('Infrared command received', extra={'command': command})
        if command == 'start':
            if not self.event_loop_thread:
                self.stop_event.clear()
                self.event_loop_thread = Thread(
                    target=functools.partial(
                        self.loop.run_until_complete,
                        self._monitor_ir(),
                    ),
                    daemon=True,
                )
                self.event_loop_thread.start()
                logger.info('Starting IR monitoring process')
            else:
                logger.info('IR monitoring process already running')
            return 'started'
        if command == 'stop':
            if self.event_loop_thread:
                self.stop_event.set()
                logger.info('Stopping IR monitoring process')
                self.event_loop_thread.join()
                self.event_loop_thread = None

            return 'stopped'
        if command == 'receive':
            logger.info('Processing receive command')
            while True:
                try:
                    code = self.ir_code_queue.get()
                except Exception:
                    logger.exception('Error retrieving IR code from queue')
                    return 'nocode'
                else:
                    logger.debug('Retrieved IR code from queue', extra={'code': code})
                    yield code
        else:
            return None

    async def _monitor_ir(self) -> None:
        """Monitor infrared signals and put received IR codes in the queue."""
        # First, find the correct IR receiver device index by checking symlinks
        logger.info('Finding IR receiver device index')

        device_index = self._get_ir_receiver_index()
        if not device_index:
            logger.error('Failed to find IR receiver device index')
            return

        logger.info('Starting ir-keytable process')
        process = await asyncio.create_subprocess_exec(
            'stdbuf',
            '-i0',
            '-o0',
            '-e0',
            'ir-keytable',
            '-c',
            '-p',
            'all',
            '-t',
            '-s',
            device_index,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.DEVNULL,
        )

        if not process.stdout:
            logger.error('Failed to start ir-keytable process - no stdout pipe')
            return

        logger.info('Started ir-keytable process with PID %d', process.pid)

        while True:
            result = await next(
                asyncio.as_completed(
                    (
                        self.stop_event.wait(),
                        process.stdout.readline(),
                    ),
                ),
            )
            if result is True:
                logger.info('Stopping IR monitoring process')
                process.kill()
                break

            line = result.decode().strip()
            logger.debug('Raw IR output', extra={'line': line})

            # Parse lirc protocol lines except repeat codes
            if not line or 'lirc protocol' not in line or 'repeat' in line:
                await asyncio.sleep(0.1)
                continue

            match = re.search(
                r'lirc protocol\((\w+)\).*scancode = (0x[0-9a-fA-F]+)',
                line,
            )
            if not match:
                logger.warning('Failed to parse IR code', extra={'line': line})
                continue

            protocol, scancode = match.group(1), match.group(2)

            logger.info(
                'IR code received',
                extra={'protocol': protocol, 'scancode': scancode},
            )
            code_str = f'{protocol}:{scancode}'
            # Put the code in the queue
            self.ir_code_queue.put(code_str)

    def _get_ir_receiver_index(self) -> str | None:
        """Get the IR receiver device index by checking symlinks."""
        try:
            # Scan /sys/class/rc/ directory for ir-receiver symlinks
            rc_dir = '/sys/class/rc'
            if Path(rc_dir).exists():
                for rc_device in Path(rc_dir).iterdir():
                    if rc_device.is_symlink():
                        link_target = Path.readlink(rc_device)
                        logger.debug(
                            'Found RC device',
                            extra={'device': rc_device, 'link_target': link_target},
                        )
                        if 'ir-receiver' in str(link_target):
                            device_name = rc_device.name
                            logger.info(
                                'Found IR receiver device index',
                                extra={
                                    'device_index': device_name,
                                    'link_target': link_target,
                                },
                            )
                            return device_name
        except OSError:
            logger.exception('Error scanning /sys/class/rc/')

        logger.error('Failed to find IR receiver device index')
        return None


infrared_manager = InfraredManager()
infrared_handler = infrared_manager.handle_command



"""gRPC service that implements the Store service."""

from __future__ import annotations

import ast
from asyncio import Queue, QueueFull
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, cast

import betterproto
from betterproto.lib.std.google import protobuf as betterproto_protobuf
from ubo_app.logger import logger
from ubo_app.rpc.message_to_object import get_class, rebuild_object, reduce_group
from ubo_app.rpc.object_to_message import GRPCSerializable, build_message
from ubo_app.store.main import RootState, UboAction, UboEvent, store
from ubo_app.utils.error_handlers import report_service_error

from ubo_bindings.store.v1 import (
    DispatchActionRequest,
    DispatchActionResponse,
    StoreServiceBase,
    SubscribeEventRequest,
    SubscribeEventResponse,
    SubscribeStoreRequest,
    SubscribeStoreResponse,
)
from ubo_bindings.ubo.v1 import Event

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Callable


def _is_valid_selector(selector: str) -> bool:
    try:
        n = ast.parse(selector, mode='eval').body
        while isinstance(n, (ast.Attribute, ast.Subscript)):
            if isinstance(n, ast.Attribute):
                n = n.value
            else:
                if not (
                    isinstance(n.slice, ast.Constant) and isinstance(n.slice.value, str)
                ):
                    return False
                n = n.value
        return isinstance(n, ast.Name) and n.id == 'state'
    except SyntaxError:
        return False


def _to_selector(selector: str) -> Callable[[RootState], Any]:
    if not _is_valid_selector(selector):
        msg = f'Invalid selector: {selector}'
        raise ValueError(msg)

    return eval(compile(f'lambda state: {selector}', '<string>', 'eval'))  # noqa: S307


def _pack_to_any(partial_state: GRPCSerializable) -> betterproto_protobuf.Any:
    """Convert a partial state to a betterproto.Message."""
    message = build_message(partial_state)

    if isinstance(message, str):
        return betterproto_protobuf.Any(
            type_url='type.googleapis.com/google.protobuf.StringValue',
            value=betterproto_protobuf.StringValue(value=message).SerializeToString(),
        )
    if isinstance(message, bytes):
        return betterproto_protobuf.Any(
            type_url='type.googleapis.com/google.protobuf.BytesValue',
            value=betterproto_protobuf.BytesValue(value=message).SerializeToString(),
        )
    if isinstance(message, bool):
        return betterproto_protobuf.Any(
            type_url='type.googleapis.com/google.protobuf.BoolValue',
            value=betterproto_protobuf.BoolValue(value=message).SerializeToString(),
        )
    if isinstance(message, int):
        return betterproto_protobuf.Any(
            type_url='type.googleapis.com/google.protobuf.Int64Value',
            value=betterproto_protobuf.Int64Value(value=message).SerializeToString(),
        )
    if isinstance(message, float):
        return betterproto_protobuf.Any(
            type_url='type.googleapis.com/google.protobuf.DoubleValue',
            value=betterproto_protobuf.DoubleValue(value=message).SerializeToString(),
        )
    if message is None:
        return betterproto_protobuf.Any(
            type_url='type.googleapis.com/google.protobuf.Empty',
            value=betterproto_protobuf.Empty().SerializeToString(),
        )
    if isinstance(message, Sequence):
        msg = 'Containers are not yet supported in the return type of a selector.'
        raise TypeError(msg)

    return betterproto_protobuf.Any(
        type_url=f'type.googleapis.com/ubo_bindings.ubo.v1.{type(message).__name__}',
        value=message.SerializeToString(),
    )


class StoreService(StoreServiceBase):
    """gRPC service class that implements the Store service."""

    async def dispatch_action(
        self,
        dispatch_action_request: DispatchActionRequest,
    ) -> DispatchActionResponse:
        """Dispatch an action to the store."""
        try:
            action = rebuild_object(dispatch_action_request.action)
        except Exception:
            logger.exception(
                'Failed to build object from dispatch action request coming from gRPC',
                extra={
                    'request': dispatch_action_request,
                },
            )
        else:
            logger.debug(
                'Dispatching action coming from gRPC',
                extra={
                    'request': dispatch_action_request,
                    'action': action,
                },
            )
            store.dispatch(cast('UboAction', action))
        return DispatchActionResponse()

    async def subscribe_event(
        self,
        subscribe_event_request: SubscribeEventRequest,
    ) -> AsyncIterator[SubscribeEventResponse]:
        """Subscribe to an event from the store."""
        logger.info(
            'Received event subscription over gRPC',
            extra={'request': subscribe_event_request},
        )
        event_class = get_class(reduce_group(subscribe_event_request.event))
        queue: Queue[UboEvent] = Queue(30)
        if event_class:

            def queue_event(event: UboEvent) -> None:
                """Put the event in the queue."""
                try:
                    queue.put_nowait(event)
                except QueueFull:
                    logger.verbose(
                        'Subscription event queue is full, dropping event',
                        extra={
                            'event': event,
                            'queue_size': queue.qsize(),
                        },
                    )

            unsubscribe = store.subscribe_event(
                event_class,
                queue_event,
                keep_ref=False,
            )
            try:
                while True:
                    event = await queue.get()
                    yield SubscribeEventResponse(
                        event=Event(
                            **{
                                betterproto.casing.snake_case(
                                    type(event).__name__,
                                ): cast('Any', build_message(event)),
                            },
                        ),
                    )
            except Exception:
                logger.exception(
                    'Exception in event subscription',
                    extra={
                        'event_type': subscribe_event_request.event,
                    },
                )
                report_service_error()
            finally:
                logger.info(
                    'Unsubscribing from event subscription over gRPC',
                    extra={'request': subscribe_event_request},
                )
                unsubscribe()

    async def subscribe_store(
        self,
        subscribe_store_request: SubscribeStoreRequest,
    ) -> AsyncIterator[SubscribeStoreResponse]:
        """Subscribe to the changes of selected parts of the store."""
        queue: Queue[Sequence[GRPCSerializable]] = Queue(30)

        selectors = [
            _to_selector(selector) for selector in subscribe_store_request.selectors
        ]

        def parent_selector(state: RootState) -> Sequence[GRPCSerializable]:
            return tuple(selector(state) for selector in selectors)

        @store.autorun(parent_selector)
        def queue_change(partial_state: Sequence[GRPCSerializable]) -> None:
            """Put the change in the queue."""
            try:
                queue.put_nowait(partial_state)
            except QueueFull:
                logger.debug(
                    'Subscription store queue is full, dropping change',
                    extra={
                        'partial_state': partial_state,
                        'queue_size': queue.qsize(),
                    },
                )

        try:
            while True:
                change = await queue.get()
                yield SubscribeStoreResponse(
                    results=[_pack_to_any(partial_state) for partial_state in change],
                )
        except Exception:
            logger.exception(
                'Exception in store subscription',
                extra={
                    'selectors': subscribe_store_request.selectors,
                },
            )
            report_service_error()
        finally:
            queue_change.unsubscribe()



# ruff: noqa: D100, D103
from __future__ import annotations

import datetime
from dataclasses import replace

from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.logger import logger
from ubo_app.store.services.assistant import (
    AssistantStartListeningAction,
    AssistantStopListeningAction,
    AssistantToggleListeningAction,
)
from ubo_app.store.services.display import (
    DisplayAction,
    DisplayBlankAction,
    DisplayBlankEvent,
    DisplayBlankTimeout,
    DisplayPauseAction,
    DisplayRedrawAction,
    DisplayRedrawEvent,
    DisplayResumeAction,
    DisplaySetBlankTimeoutAction,
    DisplayState,
    DisplayUnblankAction,
    DisplayUnblankEvent,
    DisplayUpdateActivityAction,
)

Action = (
    InitAction
    | DisplayAction
    | AssistantStartListeningAction
    | AssistantStopListeningAction
    | AssistantToggleListeningAction
)


def reducer(
    state: DisplayState | None,
    action: Action,
) -> ReducerResult[
    DisplayState,
    None,
    DisplayRedrawEvent | DisplayBlankEvent | DisplayUnblankEvent,
]:
    if state is None:
        if isinstance(action, InitAction):
            from ubo_app.utils.persistent_store import read_from_persistent_store

            logger.info('Display reducer initialized')
            selected_timeout = read_from_persistent_store(
                'display:selected_blank_timeout',
                mapper=lambda value: DisplayBlankTimeout(value),
                default=DisplayBlankTimeout.TEN_MINUTES,
            )
            return DisplayState(selected_blank_timeout=selected_timeout)
        raise InitializationActionError(action)

    # Log all display actions
    if isinstance(action, DisplayAction):
        logger.debug(
            'Display reducer received action',
            extra={'action_type': type(action).__name__},
        )

    match action:
        case DisplayPauseAction():
            return replace(state, is_paused=True)

        case DisplayResumeAction():
            return CompleteReducerResult(
                state=replace(state, is_paused=False),
                events=[DisplayRedrawEvent()],
            )

        case DisplayRedrawAction():
            return CompleteReducerResult(
                state=state,
                events=[DisplayRedrawEvent()],
            )

        case DisplayBlankAction():
            logger.info('DisplayBlankAction received in reducer')
            return CompleteReducerResult(
                state=replace(state, is_blanked=True),
                events=[DisplayBlankEvent()],
            )

        case DisplayUnblankAction():
            logger.info('DisplayUnblankAction received in reducer')
            return CompleteReducerResult(
                state=replace(
                    state,
                    is_blanked=False,
                    last_activity_time=datetime.datetime.now(tz=datetime.UTC).timestamp(),
                ),
                events=[DisplayUnblankEvent(), DisplayRedrawEvent()],
            )

        case DisplayUpdateActivityAction():
            timestamp = datetime.datetime.now(tz=datetime.UTC).timestamp()
            logger.debug(
                'DisplayUpdateActivityAction received',
                extra={'timestamp': timestamp},
            )
            return replace(state, last_activity_time=timestamp)

        case DisplaySetBlankTimeoutAction():
            logger.info(
                'DisplaySetBlankTimeoutAction received',
                extra={'timeout': action.timeout},
            )
            return replace(state, selected_blank_timeout=action.timeout)

        case (
            AssistantStartListeningAction()
            | AssistantStopListeningAction()
            | AssistantToggleListeningAction()
        ):
            timestamp = datetime.datetime.now(tz=datetime.UTC).timestamp()
            logger.info(
                'Assistant action - updating activity and waking screen if blanked',
                extra={
                    'action_type': type(action).__name__,
                    'is_blanked': state.is_blanked,
                },
            )

            if state.is_blanked:
                return CompleteReducerResult(
                    state=replace(
                        state,
                        is_blanked=False,
                        last_activity_time=timestamp,
                    ),
                    events=[DisplayUnblankEvent(), DisplayRedrawEvent()],
                )
            return replace(state, last_activity_time=timestamp)

        case _:
            return state



# ruff: noqa: D100, D103
from __future__ import annotations

from constants import AUDIO_MIC_STATE_ICON_ID, AUDIO_MIC_STATE_ICON_PRIORITY
from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.services.audio import (
    AudioAction,
    AudioChangeVolumeAction,
    AudioDevice,
    AudioEvent,
    AudioInstallDriverAction,
    AudioInstallDriverEvent,
    AudioPlayAudioSampleAction,
    AudioPlayAudioSampleEvent,
    AudioPlayAudioSequenceAction,
    AudioPlayAudioSequenceEvent,
    AudioPlaybackDoneAction,
    AudioPlaybackDoneEvent,
    AudioPlayChimeAction,
    AudioPlayChimeEvent,
    AudioPlayRecordingAction,
    AudioReportSampleAction,
    AudioReportSampleEvent,
    AudioSetMuteStatusAction,
    AudioSetVolumeAction,
    AudioStartRecordingAction,
    AudioState,
    AudioStopRecordingAction,
    AudioToggleMuteStatusAction,
    AudioToggleRecordingAction,
)
from ubo_app.store.services.notifications import Chime
from ubo_app.store.status_icons.types import StatusIconsRegisterAction

Action = InitAction | AudioAction | StatusIconsRegisterAction


def reducer(
    state: AudioState | None,
    action: Action,
) -> ReducerResult[AudioState, Action, AudioEvent]:
    if state is None:
        if isinstance(action, InitAction):
            return AudioState()
        raise InitializationActionError(action)

    match action:
        case AudioInstallDriverAction():
            return CompleteReducerResult(
                state=state,
                events=[AudioInstallDriverEvent()],
            )

        case AudioSetVolumeAction(device=AudioDevice.OUTPUT):
            return state(playback_volume=action.volume)

        case AudioSetVolumeAction(device=AudioDevice.INPUT):
            return state(capture_volume=action.volume)

        case AudioChangeVolumeAction(device=AudioDevice.OUTPUT):
            return CompleteReducerResult(
                state=state,
                actions=[
                    AudioSetVolumeAction(
                        device=AudioDevice.OUTPUT,
                        volume=min(
                            max(state.playback_volume + action.amount, 0),
                            1,
                        ),
                    ),
                ],
                events=[AudioPlayChimeEvent(name=Chime.VOLUME_CHANGE)],
            )

        case AudioChangeVolumeAction(device=AudioDevice.INPUT):
            return state(
                capture_volume=min(
                    max(state.capture_volume + action.amount, 0),
                    1,
                ),
            )

        case AudioSetMuteStatusAction(device=AudioDevice.OUTPUT):
            return state(is_playback_mute=action.is_mute)

        case AudioSetMuteStatusAction(device=AudioDevice.INPUT):
            return CompleteReducerResult(
                state=state(is_capture_mute=action.is_mute),
                actions=[
                    StatusIconsRegisterAction(
                        icon='󰍭' if action.is_mute else '󰍬',
                        priority=AUDIO_MIC_STATE_ICON_PRIORITY,
                        id=AUDIO_MIC_STATE_ICON_ID,
                    ),
                ],
            )

        case AudioToggleMuteStatusAction():
            return CompleteReducerResult(
                state=state,
                actions=[
                    AudioSetMuteStatusAction(
                        is_mute=not state.is_playback_mute
                        if action.device == AudioDevice.OUTPUT
                        else not state.is_capture_mute,
                        device=action.device,
                    ),
                ],
            )

        case AudioPlayChimeAction():
            return CompleteReducerResult(
                state=state,
                events=[
                    AudioPlayChimeEvent(name=action.name),
                ],
            )

        case AudioPlayAudioSequenceAction():
            return CompleteReducerResult(
                state=state,
                events=[
                    AudioPlayAudioSequenceEvent(
                        volume=state.playback_volume,
                        sample=action.sample,
                        id=action.id,
                        index=action.index,
                    ),
                ],
            )

        case AudioPlayAudioSampleAction():
            return CompleteReducerResult(
                state=state,
                events=[
                    AudioPlayAudioSampleEvent(
                        volume=state.playback_volume,
                        sample=action.sample,
                    ),
                ],
            )

        case AudioPlaybackDoneAction():
            return CompleteReducerResult(
                state=state,
                events=[
                    AudioPlaybackDoneEvent(id=action.id),
                ],
            )

        case AudioReportSampleAction():
            return CompleteReducerResult(
                state=state(
                    recording=action.sample()
                    if state.recording is None
                    else state.recording(
                        data=state.recording.data + action.sample.data,
                    ),
                )
                if state.is_recording
                else state,
                events=[]
                if state.is_capture_mute
                else [
                    AudioReportSampleEvent(
                        timestamp=action.timestamp,
                        sample=action.sample,
                        sample_speech_recognition=action.sample_speech_recognition,
                    ),
                ],
            )

        case AudioStartRecordingAction():
            return state(is_recording=True, recording=None)

        case AudioStopRecordingAction():
            return state(is_recording=False)

        case AudioToggleRecordingAction():
            return CompleteReducerResult(
                state=state,
                actions=[
                    AudioStopRecordingAction()
                    if state.is_recording
                    else AudioStartRecordingAction(),
                ],
            )

        case AudioPlayRecordingAction() if state.recording and not state.is_recording:
            return CompleteReducerResult(
                state=state,
                events=[
                    AudioPlayAudioSampleEvent(sample=state.recording, volume=1),
                ],
            )

        case _:
            return state



"""Vosk speech recognition engine."""

from __future__ import annotations

import asyncio
import json
from asyncio import get_event_loop
from concurrent.futures import ThreadPoolExecutor

from abstraction.speech_recognition_mixin import (
    PhraseRecognition,
    SpeechRecognition,
    SpeechRecognitionMixin,
)
from abstraction.wake_word_recognition_mixin import WakeWordRecognitionMixin
from constants import VOSK_MODEL_PATH
from typing_extensions import override

from ubo_app.constants import SPEECH_RECOGNITION_FRAME_RATE
from ubo_app.engines.vosk import VoskEngine as BaseVosk
from ubo_app.logger import logger
from ubo_app.store.main import store
from ubo_app.store.services.speech_recognition import (
    SpeechRecognitionReportTextEvent,
)
from ubo_app.utils import IS_RPI


class VoskEngine(
    BaseVosk,
    SpeechRecognitionMixin,
    WakeWordRecognitionMixin,
):
    """Vosk speech recognition engine."""

    def __init__(self) -> None:
        """Initialize Vosk speech recognition engine."""
        self.grammar_lock = asyncio.Lock()
        self.process_executor = ThreadPoolExecutor(max_workers=1)

        super().__init__(label='Vosk')

    @override
    async def _run(self) -> None:
        from vosk import KaldiRecognizer, Model

        phrases = self._phrases
        model = Model(
            model_path=VOSK_MODEL_PATH.resolve().as_posix(),
            lang='en-us',
        )
        logger.debug(
            'Vosk - Starting recognition loop',
            extra={
                'engine_name': self.name,
                'phrases': phrases,
            },
        )
        recognizer = KaldiRecognizer(
            model,
            SPEECH_RECOGNITION_FRAME_RATE,
            *([json.dumps(phrases)] if phrases else []),
        )

        while self.should_be_running():
            data = await self.input_queue.get()

            if await get_event_loop().run_in_executor(
                self.process_executor,
                recognizer.AcceptWaveform,
                data,
            ):
                result = json.loads(recognizer.FinalResult())

                if result.get('text'):
                    await self.report(result=result['text'])
            else:
                result = json.loads(recognizer.PartialResult())
                if result.get('partial'):
                    logger.verbose(
                        'Vosk - Partial result',
                        extra={'result': result},
                    )
                    store._dispatch(  # noqa: SLF001
                        [
                            SpeechRecognitionReportTextEvent(
                                timestamp=get_event_loop().time(),
                                text=result['partial'],
                            ),
                        ],
                    )

            if self.ongoing_recognition is not None:
                self.ongoing_recognition.append_voice(data)

            async with self.grammar_lock:
                if (_phrases := self._phrases) != phrases:
                    phrases = _phrases
                    logger.debug(
                        'Vosk - Updating phrases',
                        extra={
                            'new_phrases': phrases,
                        },
                    )
                    if IS_RPI:
                        recognizer.Reset()
                        recognizer.SetGrammar(json.dumps(phrases))
                    else:
                        recognizer = KaldiRecognizer(
                            model,
                            SPEECH_RECOGNITION_FRAME_RATE,
                            *([json.dumps(phrases)] if phrases else []),
                        )

    @property
    def _phrases(self) -> tuple[str, ...] | None:
        """Get the phrases for the Vosk recognizer."""
        if self.ongoing_recognition:
            if isinstance(self.ongoing_recognition, SpeechRecognition):
                return ()
            if isinstance(self.ongoing_recognition, PhraseRecognition):
                return (
                    *self.ongoing_recognition.phrases,
                    '[unk]',
                )
            msg = 'Ongoing recognition must have either end_phrase or phrases set.'
            raise ValueError(msg)

        if self.wake_words:
            return (*self.wake_words, '[unk]')

        return None



# ruff: noqa: D100, D101
from __future__ import annotations

from enum import StrEnum
from typing import TYPE_CHECKING

from immutable import Immutable
from redux import BaseAction, BaseEvent

from ubo_app.utils.persistent_store import read_from_persistent_store


class CameraType(StrEnum):
    """Camera type enum."""

    DEFAULT = 'default'
    AUTOFOCUS = 'autofocus'
    FIXED_FOCUS = 'fixed-focus'

if TYPE_CHECKING:
    from ubo_app.store.input.types import QRCodeInputDescription


class CameraAction(BaseAction): ...


class CameraStartViewfinderAction(CameraAction):
    pattern: str | None


class CameraReportBarcodeAction(CameraAction):
    codes: list[str]


class CameraInstallDriverAction(CameraAction):
    """Install camera driver action."""

    make: str
    model: str
    variant: str


class CameraEvent(BaseEvent): ...


class CameraInstallDriverEvent(CameraEvent):
    """Install camera driver event."""

    make: str
    model: str
    variant: str


class CameraRestoreDefaultAction(CameraAction):
    """Restore default camera configuration action."""


class CameraRestoreDefaultEvent(CameraEvent):
    """Restore default camera configuration event."""


class CameraStartViewfinderEvent(CameraEvent):
    pattern: str | None


class CameraReportImageEvent(CameraEvent):
    """Event for reporting an image from the camera."""

    timestamp: float
    data: bytes
    width: int
    height: int


class CameraStopViewfinderEvent(CameraEvent): ...


class CameraSetIndexAction(CameraAction):
    """Action to set the selected camera index."""

    index: int


class CameraDetectAction(CameraAction):
    """Action to trigger camera detection."""


class CameraSetAvailableCamerasAction(CameraAction):
    """Action to set available cameras."""

    available_cameras: list[int]


class CameraDetectEvent(CameraEvent):
    """Event fired to trigger camera detection."""


class CameraDetectedEvent(CameraEvent):
    """Event fired when cameras are detected."""

    available_cameras: list[int]


class CameraReinitializeEvent(CameraEvent):
    """Event to trigger camera reinitialization with new index."""


class CameraState(Immutable):
    queue: list[QRCodeInputDescription]
    selected_camera_index: int = 0
    available_cameras: tuple[int, ...] = ()
    camera_type: CameraType = read_from_persistent_store(
        'camera_type',
        default=CameraType.DEFAULT,
    )



# ruff: noqa: D100, D103
from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Literal

from kivy.core.window import Keyboard, Window, WindowBase

from ubo_app.store.services.assistant import (
    AssistantStartListeningAction,
    AssistantStopListeningAction,
)
from ubo_app.store.services.audio import AudioDevice, AudioToggleMuteStatusAction
from ubo_app.store.services.keypad import (
    Key,
    KeypadKeyPressAction,
    KeypadKeyReleaseAction,
)

if TYPE_CHECKING:
    from ubo_app.utils.types import Subscriptions

    Modifier = Literal['ctrl', 'alt', 'meta', 'shift']

KEY_MAP = {
    'no_modifier': {
        Keyboard.keycodes['up']: KeypadKeyPressAction(
            key=Key.UP,
            pressed_keys={Key.UP},
        ),
        Keyboard.keycodes['k']: KeypadKeyPressAction(key=Key.UP, pressed_keys={Key.UP}),
        Keyboard.keycodes['down']: KeypadKeyPressAction(
            key=Key.DOWN,
            pressed_keys={Key.DOWN},
        ),
        Keyboard.keycodes['j']: KeypadKeyPressAction(
            key=Key.DOWN,
            pressed_keys={Key.DOWN},
        ),
        Keyboard.keycodes['1']: KeypadKeyPressAction(key=Key.L1, pressed_keys={Key.L1}),
        Keyboard.keycodes['2']: KeypadKeyPressAction(key=Key.L2, pressed_keys={Key.L2}),
        Keyboard.keycodes['3']: KeypadKeyPressAction(key=Key.L3, pressed_keys={Key.L3}),
        Keyboard.keycodes['left']: KeypadKeyReleaseAction(
            key=Key.BACK,
            pressed_keys=set(),
        ),
        Keyboard.keycodes['escape']: KeypadKeyReleaseAction(
            key=Key.BACK,
            pressed_keys=set(),
        ),
        Keyboard.keycodes['h']: KeypadKeyReleaseAction(
            key=Key.BACK,
            pressed_keys=set(),
        ),
        Keyboard.keycodes['backspace']: KeypadKeyReleaseAction(
            key=Key.HOME,
            pressed_keys=set(),
        ),
        Keyboard.keycodes['m']: AudioToggleMuteStatusAction(device=AudioDevice.INPUT),
    },
    'ctrl': {
        Keyboard.keycodes['up']: KeypadKeyPressAction(
            key=Key.UP,
            pressed_keys={Key.BACK, Key.UP},
        ),
        Keyboard.keycodes['k']: KeypadKeyPressAction(
            key=Key.UP,
            pressed_keys={Key.BACK, Key.UP},
        ),
        Keyboard.keycodes['down']: KeypadKeyPressAction(
            key=Key.DOWN,
            pressed_keys={Key.BACK, Key.DOWN},
        ),
        Keyboard.keycodes['j']: KeypadKeyPressAction(
            key=Key.DOWN,
            pressed_keys={Key.BACK, Key.DOWN},
        ),
        Keyboard.keycodes['1']: KeypadKeyPressAction(
            key=Key.L1,
            pressed_keys={Key.BACK, Key.L1},
        ),
        Keyboard.keycodes['2']: KeypadKeyPressAction(
            key=Key.L2,
            pressed_keys={Key.BACK, Key.L2},
        ),
        Keyboard.keycodes['3']: KeypadKeyPressAction(
            key=Key.L3,
            pressed_keys={Key.BACK, Key.L3},
        ),
        Keyboard.keycodes['backspace']: KeypadKeyPressAction(
            key=Key.HOME,
            pressed_keys={Key.BACK, Key.HOME},
        ),
    },
    'shift': {
        Keyboard.keycodes['up']: KeypadKeyPressAction(
            key=Key.UP,
            pressed_keys={Key.HOME, Key.UP},
        ),
        Keyboard.keycodes['k']: KeypadKeyPressAction(
            key=Key.UP,
            pressed_keys={Key.HOME, Key.UP},
        ),
        Keyboard.keycodes['down']: KeypadKeyPressAction(
            key=Key.DOWN,
            pressed_keys={Key.HOME, Key.DOWN},
        ),
        Keyboard.keycodes['j']: KeypadKeyPressAction(
            key=Key.DOWN,
            pressed_keys={Key.HOME, Key.DOWN},
        ),
        Keyboard.keycodes['1']: KeypadKeyPressAction(
            key=Key.L1,
            pressed_keys={Key.HOME, Key.L1},
        ),
        Keyboard.keycodes['2']: KeypadKeyPressAction(
            key=Key.L2,
            pressed_keys={Key.HOME, Key.L2},
        ),
        Keyboard.keycodes['3']: KeypadKeyPressAction(
            key=Key.L3,
            pressed_keys={Key.HOME, Key.L3},
        ),
        Keyboard.keycodes['left']: KeypadKeyPressAction(
            key=Key.BACK,
            pressed_keys={Key.HOME, Key.BACK},
        ),
        Keyboard.keycodes['escape']: KeypadKeyPressAction(
            key=Key.BACK,
            pressed_keys={Key.HOME, Key.BACK},
        ),
        Keyboard.keycodes['h']: KeypadKeyPressAction(
            key=Key.BACK,
            pressed_keys={Key.HOME, Key.BACK},
        ),
    },
}


def on_keyboard(
    window: WindowBase,
    key: int,
    scancode: int,
    codepoint: str,
    modifier: list[Modifier],
) -> None:
    """Handle keyboard events."""
    _ = window, scancode, codepoint
    from ubo_app.store.main import store

    if modifier == [] and key in KEY_MAP['no_modifier']:
        store.dispatch(KEY_MAP['no_modifier'][key])
    elif modifier == ['ctrl'] and key in KEY_MAP['ctrl']:
        store.dispatch(KEY_MAP['ctrl'][key])
    elif modifier == ['shift'] and key in KEY_MAP['shift']:
        store.dispatch(KEY_MAP['shift'][key])


def on_key_down(
    window: WindowBase,
    key: int,
    scancode: int,
    codepoint: str,
    modifier: list[Modifier],
) -> None:
    """Handle key down events."""
    _ = window, scancode, codepoint
    from ubo_app.store.main import store

    if modifier == [] and key == Keyboard.keycodes['v']:
        store.dispatch(AssistantStartListeningAction())


def on_key_up(
    window: WindowBase,
    key: int,
    scancode: int,
) -> None:
    """Handle key up events."""
    _ = window, scancode
    from ubo_app.store.main import store

    if key == Keyboard.keycodes['v']:
        store.dispatch(AssistantStopListeningAction())


def init_service() -> Subscriptions:
    Window.bind(on_keyboard=on_keyboard)
    Window.bind(on_key_down=on_key_down)
    Window.bind(on_key_up=on_key_up)

    return [
        functools.partial(Window.unbind, on_keyboard=on_keyboard),
        functools.partial(Window.unbind, on_key_down=on_key_down),
        functools.partial(Window.unbind, on_key_up=on_key_up),
    ]



# ruff: noqa: D100, D103
from __future__ import annotations

from dataclasses import replace

from redux import CompleteReducerResult, InitAction, InitializationActionError

from ubo_app.store.services.speech_synthesis import (
    SpeechSynthesisAction,
    SpeechSynthesisEvent,
    SpeechSynthesisReadTextAction,
    SpeechSynthesisSetSelectedEngineAction,
    SpeechSynthesisState,
    SpeechSynthesisSynthesizeTextEvent,
    SpeechSynthesisUpdateAccessKeyStatus,
)


def reducer(
    state: SpeechSynthesisState | None,
    action: SpeechSynthesisAction,
) -> (
    SpeechSynthesisState
    | CompleteReducerResult[
        SpeechSynthesisState,
        SpeechSynthesisAction,
        SpeechSynthesisEvent,
    ]
):
    if state is None:
        if isinstance(action, InitAction):
            return SpeechSynthesisState()
        raise InitializationActionError(action)

    match action:
        case SpeechSynthesisUpdateAccessKeyStatus():
            return replace(state, is_access_key_set=action.is_access_key_set)

        case SpeechSynthesisSetSelectedEngineAction():
            return replace(state, selected_engine=action.engine_name)

        case SpeechSynthesisReadTextAction():
            return CompleteReducerResult(
                state=state,
                events=[
                    SpeechSynthesisSynthesizeTextEvent(
                        information=action.information,
                        speech_rate=action.speech_rate,
                    ),
                ],
            )

        case _:
            return state



# ruff: noqa: D100, D103
from __future__ import annotations

from dataclasses import replace

from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.services.rgb_ring import (
    RgbRingAction,
    RgbRingCommandAction,
    RgbRingCommandEvent,
    RgbRingSetIsBusyAction,
    RgbRingState,
)

Action = InitAction | RgbRingAction


def reducer(
    state: RgbRingState | None,
    action: Action,
) -> ReducerResult[RgbRingState, Action, RgbRingCommandEvent]:
    if state is None:
        if isinstance(action, InitAction):
            return RgbRingState(is_busy=False)
        raise InitializationActionError(action)

    match action:
        case RgbRingSetIsBusyAction():
            return replace(
                state,
                is_busy=action.is_busy,
            )

        case RgbRingCommandAction():
            command = action.as_command()
            if not command:
                return state
            return CompleteReducerResult(
                state=state,
                events=[RgbRingCommandEvent(command=command.split())],
            )

        case _:
            return state



# pyright: reportMissingModuleSource=false
"""Provides the system bus for the application."""

from __future__ import annotations

from threading import Thread, current_thread

from sdbus import SdBus, sd_bus_open_system, sd_bus_open_user, set_default_bus

system_buses: dict[Thread, SdBus] = {}
user_buses: dict[Thread, SdBus] = {}


def clean_up() -> None:
    """Cleanup the buses."""
    for bus in system_buses.values():
        bus.close()
    system_buses.clear()
    for bus in user_buses.values():
        bus.close()
    user_buses.clear()


def get_system_bus() -> SdBus:
    """Get the system bus for current thread."""
    thread = current_thread()
    if thread not in system_buses:
        system_buses[thread] = sd_bus_open_system()
    set_default_bus(system_buses[thread])
    return system_buses[thread]


def get_user_bus() -> SdBus:
    """Get the system bus for current thread."""
    thread = current_thread()
    if thread not in user_buses:
        user_buses[thread] = sd_bus_open_user()
    return user_buses[thread]



# pyright: reportMissingModuleSource=false
"""Module for managing audio playback and recording."""

from __future__ import annotations

import asyncio
import math
import wave
from asyncio import get_event_loop
from concurrent.futures import ThreadPoolExecutor
from threading import Lock
from typing import TYPE_CHECKING

import alsaaudio
import numpy as np
import simpleaudio
import soxr
from simpleaudio import _simpleaudio  # pyright: ignore [reportAttributeAccessIssue]
from tenacity import (
    AsyncRetrying,
    retry_if_exception,
    stop_after_attempt,
    wait_fixed,
)

from ubo_app.constants import SPEECH_RECOGNITION_FRAME_RATE
from ubo_app.logger import logger
from ubo_app.store.main import store
from ubo_app.store.services.audio import (
    AudioPlaybackDoneAction,
    AudioReportSampleAction,
    AudioSample,
)
from ubo_app.utils import IS_RPI
from ubo_app.utils.async_ import create_task
from ubo_app.utils.eeprom import get_eeprom_data
from ubo_app.utils.error_handlers import report_service_error
from ubo_app.utils.server import send_command

if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine

INPUT_FRAME_RATE = 48_000
INPUT_CHANNELS = 2
INPUT_PERIOD_SIZE = int(INPUT_FRAME_RATE / 1000) * 20  # 20ms


def _linear_to_logarithmic(volume_linear: float) -> int:
    """Convert a linear volume to a logarithmic volume.

    Assuming volume_linear is between 0 and 1
    Convert it to a logarithmic scale.
    """
    if volume_linear == 0:
        return 0
    return round(100 * math.log(volume_linear * 500) / math.log(500))


class AudioManager:
    """Class for managing audio playback and recording."""

    def __init__(self) -> None:
        """Initialize the audio manager."""
        # create an audio object
        self.has_speakers = False
        self.has_microphones = False

        self.playback_mute = True
        self.playback_volume = 0.1
        self.capture_volume = 0.1

        self.audio_buffers: dict[str, dict[int, AudioSample | None]] = {}
        self.audio_heads: dict[str, int] = {}
        self.audio_buffers_lock = Lock()

        eeprom_data = get_eeprom_data()

        if (
            (speakers := eeprom_data.get('speakers'))
            and speakers.get('model') == 'wm8960'
        ):
            self.has_speakers = True

        if (
            (microphones := eeprom_data.get('microphones'))
            and microphones.get('model') == 'wm8960'
        ):
            self.has_microphones = True

        create_task(self.find_card_index())
        create_task(self.stream_mic())

        if not IS_RPI:
            import pyaudio

            self.pa = pyaudio.PyAudio()
        else:
            self.pa = None

    def close(self) -> None:
        """Close the audio manager."""
        # Close the audio buffers
        with self.audio_buffers_lock:
            self.audio_buffers.clear()
            self.audio_heads.clear()

    async def find_card_index(self) -> None:
        """Find the card index of the audio device."""
        self.card_index = None
        if not self.has_speakers and not self.has_microphones:
            return
        async for attempt in AsyncRetrying(
            stop=stop_after_attempt(3),
            wait=wait_fixed(1),
            retry=retry_if_exception(lambda e: isinstance(e, StopIteration)),
        ):
            with attempt:
                # Get the card index of the audio device
                cards = alsaaudio.cards()
                self.card_index = cards.index(
                    next(card for card in cards if 'wm8960' in card),
                )
            if attempt.retry_state.outcome and isinstance(
                attempt.retry_state.outcome.exception(),
                StopIteration,
            ):
                report_service_error(exception=attempt.retry_state.outcome.exception())
                await send_command('audio', 'failure_report', has_output=True)
            else:
                break
        else:
            logger.error(
                'Audio - Failed to find the card index after multiple trials',
            )
            return
        # In case they were set before the card was initialized
        self.set_playback_mute(mute=self.playback_mute)
        self.set_playback_volume(self.playback_volume)
        self.set_capture_volume(self.capture_volume)

    async def play_file(self, filename: str) -> None:
        """Play a waveform audio file.

        Parameters
        ----------
        filename: str
            Path to wav file

        """
        # open the file for reading.
        logger.info(
            'Audio - Opening audio file for playback',
            extra={'filename_': filename},
        )
        with wave.open(filename, 'rb') as wave_file:
            sample_rate = wave_file.getframerate()
            channels = wave_file.getnchannels()
            sample_width = wave_file.getsampwidth()
            audio_data = wave_file.readframes(wave_file.getnframes())

            await self.play_sample(
                AudioSample(
                    data=audio_data,
                    channels=channels,
                    rate=sample_rate,
                    width=sample_width,
                ),
            )

    async def play_sample(
        self,
        sample: AudioSample,
    ) -> None:
        """Play an audio sample.

        Parameters
        ----------
        sample: AudioSample
            Audio sample as a sequence of bytes and its parameters: sample rate, width
            and channels

        """
        if sample.data == b'':
            return
        async for attempt in AsyncRetrying(
            stop=stop_after_attempt(3),
            wait=wait_fixed(1),
        ):
            with attempt:
                play_object = simpleaudio.play_buffer(
                    audio_data=sample.data,
                    num_channels=sample.channels,
                    sample_rate=sample.rate,
                    bytes_per_sample=sample.width,
                )
                play_object.wait_done()
            if attempt.retry_state.outcome and attempt.retry_state.outcome.exception():
                if isinstance(
                    attempt.retry_state.outcome.exception(),
                    _simpleaudio.SimpleaudioError,
                ):
                    logger.info(
                        'Audio - Reporting the playback issue to ubo-system',
                        extra={'attempt': attempt.retry_state.attempt_number},
                    )
                    report_service_error(
                        exception=attempt.retry_state.outcome.exception(),
                    )
                    await send_command('audio', 'failure_report', has_output=True)
            else:
                break
        else:
            logger.error(
                'Audio - Failed to play sample after multiple trials',
            )
            return

    async def play_sequence(  # noqa: C901
        self,
        sample: AudioSample | None,
        *,
        id: str,
        index: int,
    ) -> None:
        """Play a sequence of audio samples.

        Parameters
        ----------
        sample: AudioSample
            Audio sample as a sequence of bytes and its parameters: sample rate, width
            and channels

        id: str
            ID of the audio sequence chain

        index: int
            Index of the sample in the sequence

        """
        with self.audio_buffers_lock:
            if not (already_playing := id in self.audio_buffers):
                self.audio_buffers[id] = {}
                self.audio_heads[id] = 0

        buffer = self.audio_buffers[id]
        buffer[index] = sample

        if already_playing or sample is None:
            return

        class NotProvided: ...

        not_provided = NotProvided()

        if self.pa:
            default_info = self.pa.get_default_output_device_info()
            default_playback_index = default_info['index']
            if not isinstance(default_playback_index, int):
                msg = 'Default output device index is not an integer'
                raise RuntimeError(msg)

            stream = self.pa.open(
                format=self.pa.get_format_from_width(sample.width),
                channels=sample.channels,
                rate=sample.rate,
                output=True,
                frames_per_buffer=1024,
                output_device_index=default_playback_index,
            )

            async def play(sample: AudioSample) -> None:
                """Play a sample using PyAudio."""
                chunk_size = 1024 * sample.channels * sample.width
                for i in range(0, len(sample.data), chunk_size):
                    chunk = sample.data[i : i + chunk_size]
                    stream.write(chunk)
        else:
            stream = alsaaudio.PCM(
                type=alsaaudio.PCM_PLAYBACK,
                mode=alsaaudio.PCM_NORMAL,
                channels=sample.channels,
                rate=sample.rate,
                format=alsaaudio.PCM_FORMAT_S16_LE,
                periodsize=len(sample.data) // (2 * sample.width),
            )

            async def play(sample: AudioSample) -> None:
                stream.write(sample.data)

        while (
            id in self.audio_heads
            and (head_sample := buffer.get(self.audio_heads[id], not_provided))
            is not None
        ):
            if isinstance(head_sample, NotProvided):
                await asyncio.sleep(0.05)
                continue
            await play(head_sample)
            del buffer[self.audio_heads[id]]
            self.audio_heads[id] += 1

        with self.audio_buffers_lock:
            del self.audio_buffers[id]
            del self.audio_heads[id]
            store.dispatch(AudioPlaybackDoneAction(id=id))

        if self.pa:
            import pyaudio

            if isinstance(stream, pyaudio.Stream):
                stream.stop_stream()
        stream.close()

    async def _initialize_input_reader(  # noqa: C901
        self,
    ) -> Callable[[], Coroutine[None, None, tuple[int, bytes, int]]]:
        read_executor = ThreadPoolExecutor(max_workers=1)

        if self.pa:
            import pyaudio

            try:
                channels = self.pa.get_default_input_device_info()['maxInputChannels']
                if not isinstance(channels, int) or channels < 1:

                    async def read_audio_chunk() -> tuple[int, bytes, int]:
                        await asyncio.sleep(0.1)
                        return 0, b'', 1
                else:
                    input_audio = self.pa.open(
                        format=pyaudio.paInt16,
                        channels=channels,
                        rate=INPUT_FRAME_RATE,
                        input=True,
                        frames_per_buffer=INPUT_PERIOD_SIZE,
                    )

                    async def read_audio_chunk() -> tuple[int, bytes, int]:
                        data = await get_event_loop().run_in_executor(
                            read_executor,
                            input_audio.read,
                            INPUT_PERIOD_SIZE,
                            False,  # noqa: FBT003
                        )
                        return len(data), data, channels
            except OSError:
                logger.exception('Audio - Error opening audio capture')

                async def read_audio_chunk() -> tuple[int, bytes, int]:
                    await asyncio.sleep(0.1)
                    return 0, b'', 1
        else:
            import alsaaudio  # type: ignore [reportMissingModuleSource=false]

            async for attempt in AsyncRetrying(
                stop=stop_after_attempt(3),
                wait=wait_fixed(1),
                reraise=True,
            ):
                with attempt:
                    if self.card_index is None:
                        msg = 'Card index is not set'
                        raise RuntimeError(msg)
                    input_audio = alsaaudio.PCM(
                        alsaaudio.PCM_CAPTURE,
                        alsaaudio.PCM_NORMAL,
                        channels=INPUT_CHANNELS,
                        rate=INPUT_FRAME_RATE,
                        format=alsaaudio.PCM_FORMAT_S16_LE,
                        periodsize=INPUT_PERIOD_SIZE,
                        cardindex=self.card_index,
                    )
                if attempt.retry_state.outcome and isinstance(
                    attempt.retry_state.outcome.exception(),
                    Exception,
                ):
                    if isinstance(
                        attempt.retry_state.outcome.exception(),
                        alsaaudio.ALSAAudioError,
                    ):
                        logger.info(
                            'Audio - Reporting the audio capture issue to ubo-system',
                            extra={'attempt': attempt.retry_state.attempt_number},
                        )
                        report_service_error(
                            exception=attempt.retry_state.outcome.exception(),
                        )
                        await send_command('audio', 'failure_report', has_output=True)
                    continue
                break
            else:
                # Since reraise is set to True, this part should be unreachable
                logger.error(
                    'Audio - Failed to open audio capture after multiple trials',
                )
                msg = 'Failed to open audio capture after multiple trials'
                raise RuntimeError(msg)

            async def read_audio_chunk() -> tuple[int, bytes, int]:
                result = await get_event_loop().run_in_executor(
                    read_executor,
                    input_audio.read,
                )
                return (*result, INPUT_CHANNELS)

        return read_audio_chunk

    async def stream_mic(self) -> None:
        """Stream audio from the microphone to the store."""
        read_audio_chunk = await self._initialize_input_reader()
        event_loop = get_event_loop()

        while True:
            try:
                length, data, channels = await read_audio_chunk()
            except alsaaudio.ALSAAudioError:
                logger.exception('Audio - Error reading audio capture')
                read_audio_chunk = await self._initialize_input_reader()
                break
            else:
                if length > 0:
                    data_speech_recognition = np.frombuffer(data, dtype=np.int16)
                    data_speech_recognition = data_speech_recognition.reshape(
                        -1,
                        channels,
                    )
                    data_speech_recognition = data_speech_recognition.T
                    data_speech_recognition = (
                        data_speech_recognition.astype(np.float32) / 32768.0
                    )

                    data_speech_recognition = (
                        data_speech_recognition.squeeze()
                        if channels == 1
                        else np.mean(data_speech_recognition, axis=0)
                    )

                    if INPUT_FRAME_RATE != SPEECH_RECOGNITION_FRAME_RATE:
                        data_speech_recognition = soxr.resample(
                            data_speech_recognition,
                            in_rate=INPUT_FRAME_RATE,
                            out_rate=SPEECH_RECOGNITION_FRAME_RATE,
                        )

                    data_speech_recognition = (
                        (data_speech_recognition * 32768.0).astype(np.int16).tobytes()
                    )
                    store.dispatch(
                        AudioReportSampleAction(
                            timestamp=event_loop.time(),
                            sample_speech_recognition=data_speech_recognition,
                            sample=AudioSample(
                                data=data,
                                channels=channels,
                                rate=INPUT_FRAME_RATE,
                                width=2,
                            ),
                        ),
                    )

    def set_playback_mute(self, *, mute: bool = False) -> None:
        """Set the playback mute of the audio output.

        Parameters
        ----------
        mute: bool
            Mute to set

        """
        self.playback_mute = mute
        try:
            # Assume pulseaudio is installed
            mixer = alsaaudio.Mixer(control='Master')
            mixer.setmute(1 if mute else 0)
        except alsaaudio.ALSAAudioError:
            # Seems like pulseaudio is not installed, so we directly use device mixers
            if self.card_index is None or not self.has_speakers:
                return

            try:
                mixer = alsaaudio.Mixer(
                    control='Right Output Mixer PCM',
                    cardindex=self.card_index,
                )
                mixer.setmute(0)
                mixer = alsaaudio.Mixer(
                    control='Left Output Mixer PCM',
                    cardindex=self.card_index,
                )
                mixer.setmute(0)
            except alsaaudio.ALSAAudioError:
                create_task(self.find_card_index())

    def set_playback_volume(self, volume: float = 0.8) -> None:
        """Set the playback volume of the audio output.

        Parameters
        ----------
        volume: float
            Volume to set, a float between 0 and 1

        """
        if volume < 0 or volume > 1:
            msg = 'Volume must be between 0 and 1'
            raise ValueError(msg)
        self.playback_volume = volume
        try:
            # Assume pulseaudio is installed
            mixer = alsaaudio.Mixer(control='Master')
            mixer.setvolume(round(volume * 100))
        except alsaaudio.ALSAAudioError:
            # Seems like pulseaudio is not installed, so we directly use device mixers
            if self.card_index is None or not self.has_speakers:
                return

            try:
                mixer = alsaaudio.Mixer(control='Speaker', cardindex=self.card_index)
                mixer.setvolume(
                    _linear_to_logarithmic(volume),
                    alsaaudio.MIXER_CHANNEL_ALL,
                    alsaaudio.PCM_PLAYBACK,
                )
                mixer = alsaaudio.Mixer(control='Playback', cardindex=self.card_index)
                mixer.setvolume(
                    100,
                    alsaaudio.MIXER_CHANNEL_ALL,
                    alsaaudio.PCM_PLAYBACK,
                )
            except alsaaudio.ALSAAudioError:
                create_task(self.find_card_index())

    def set_capture_volume(self, volume: float = 0.8) -> None:
        """Set the capture volume of the audio output.

        Parameters
        ----------
        volume: float
            Volume to set, a float between 0 and 1

        """
        if volume < 0 or volume > 1:
            msg = 'Volume must be between 0 and 1'
            raise ValueError(msg)
        self.capture_volume = volume
        try:
            # Assume pulseaudio is installed
            mixer = alsaaudio.Mixer(control='Capture')
            mixer.setrec(round(volume * 100))
        except alsaaudio.ALSAAudioError:
            # Seems like pulseaudio is not installed, so we directly use device mixers
            if self.card_index is None or not self.has_microphones:
                return

            try:
                mixer = alsaaudio.Mixer(control='Capture', cardindex=self.card_index)
                mixer.setvolume(
                    _linear_to_logarithmic(volume),
                    alsaaudio.MIXER_CHANNEL_ALL,
                    alsaaudio.PCM_CAPTURE,
                )
            except alsaaudio.ALSAAudioError:
                create_task(self.find_card_index())



# ruff: noqa: D100, D101, D102, D103, D107
from __future__ import annotations

import asyncio
import contextlib
import threading
import traceback
from typing import TYPE_CHECKING, TypeVarTuple

from typing_extensions import TypeVar

if TYPE_CHECKING:
    from pathlib import Path

service_id: str
service_uid: str
name: str
label: str
path: Path

if TYPE_CHECKING:
    from asyncio import Handle
    from collections.abc import Coroutine

    from redux.basic_types import TaskCreatorCallback


T = TypeVar('T', infer_variance=True)
Ts = TypeVarTuple('Ts')


class WorkerThread(threading.Thread):
    loop: asyncio.AbstractEventLoop

    def __init__(self: WorkerThread) -> None:
        super().__init__()
        self.name = 'Worker Thread'

        self.is_finished = threading.Event()
        self.is_started = threading.Event()

    def run(self: WorkerThread) -> None:
        if not self.loop.is_running():
            self.is_started.set()
            self.loop.run_forever()

    def run_coroutine(
        self: WorkerThread,
        coroutine: Coroutine,
        callback: TaskCreatorCallback | None = None,
        name: str | None = None,
    ) -> Handle:
        from ubo_app.constants import DEBUG_TASKS

        def task_wrapper(stack: str) -> None:
            task = self.loop.create_task(coroutine, name=name)
            if DEBUG_TASKS:
                from ubo_app.utils.error_handlers import STACKS

                STACKS[task] = stack
            if callback:
                callback(task)

        task_wrapper.__name__ = f'run_task:{coroutine.__name__}'
        task_wrapper.__qualname__ = f'run_task:{coroutine.__qualname__}'

        return self.loop.call_soon_threadsafe(
            task_wrapper,
            ''.join(traceback.format_stack()[:-3]) if DEBUG_TASKS else '',
        )

    async def shutdown(self: WorkerThread) -> None:
        from ubo_app.constants import MAIN_LOOP_GRACE_PERIOD
        from ubo_app.logger import logger

        logger.info('Stopping the worker thread')

        while True:
            tasks = [
                task
                for task in asyncio.all_tasks(self.loop)
                if task is not asyncio.current_task(self.loop)
                and task.cancelling() == 0
                and not task.done()
            ]
            logger.debug(
                'Waiting for tasks to finish',
                extra={
                    'tasks': tasks,
                    'thread_': self,
                },
            )
            if not tasks:
                break
            with contextlib.suppress(BaseException):
                await asyncio.wait_for(
                    asyncio.gather(
                        *tasks,
                        return_exceptions=True,
                    ),
                    timeout=MAIN_LOOP_GRACE_PERIOD,
                )
            await asyncio.sleep(0.1)

        logger.debug('Stopping event loop of the worker thread')
        self.loop.stop()
        self.is_finished.set()
        logger.info('The worker thread is done')

    def stop(self: WorkerThread) -> None:
        self.loop.call_soon_threadsafe(self.loop.create_task, self.shutdown())


worker_thread = WorkerThread()


def start_event_loop_thread(loop: asyncio.AbstractEventLoop) -> None:
    from ubo_app.utils.error_handlers import loop_exception_handler

    loop.set_exception_handler(loop_exception_handler)

    worker_thread.loop = loop
    worker_thread.start()

    from redux.basic_types import FinishEvent

    from ubo_app.store.main import store

    store.subscribe_event(FinishEvent, worker_thread.stop)

    worker_thread.is_started.wait()


def run_coroutine(
    coroutine: Coroutine,
    callback: TaskCreatorCallback | None = None,
    name: str | None = None,
) -> Handle:
    return worker_thread.run_coroutine(coroutine, callback, name=name)



# pyright: reportMissingModuleSource=false
# ruff: noqa: D100, D103
from __future__ import annotations

import asyncio
import contextlib
import uuid
from typing import TYPE_CHECKING, Any, TypeVar, cast

from debouncer import DebounceOptions, debounce
from tenacity import AsyncRetrying, stop_after_attempt, wait_fixed

from ubo_app.colors import DANGER_COLOR
from ubo_app.logger import logger
from ubo_app.store.main import store
from ubo_app.store.services.ethernet import NetState
from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.wifi import ConnectionState, WiFiConnection, WiFiType
from ubo_app.utils.bus_provider import get_system_bus
from ubo_app.utils.error_handlers import report_service_error

if TYPE_CHECKING:
    from asyncio.tasks import _FutureLike
    from collections.abc import Coroutine

from sdbus import dbus_exceptions
from sdbus.utils.inspect import (  # pyright: ignore [reportMissingImports]
    inspect_dbus_path,
)
from sdbus_async.networkmanager import (
    AccessPoint,
    ActiveConnection,
    DeviceState,
    NetworkConnectionSettings,
    NetworkDeviceGeneric,
    NetworkDeviceWireless,
    NetworkManager,
    NetworkManagerConnectionProperties,
    NetworkManagerSettings,
)
from sdbus_async.networkmanager.enums import (
    ConnectionState as SdBusConnectionState,
)
from sdbus_async.networkmanager.enums import DeviceType

RETRIES = 3

T = TypeVar('T')


class WiFiDeviceNotAvailableError(Exception):
    """Exception raised when WiFi device is not yet available."""


def wait_for(task: _FutureLike[T]) -> Coroutine[Any, Any, T]:
    return asyncio.wait_for(task, timeout=10.0)


async def get_wifi_device() -> NetworkDeviceWireless | None:
    network_manager = NetworkManager(get_system_bus())
    devices_paths = await wait_for(network_manager.get_devices())
    for device_path in devices_paths:
        generic_device = NetworkDeviceGeneric(device_path, get_system_bus())
        if (
            await wait_for(
                generic_device.device_type,
            )
            == DeviceType.WIFI
        ):
            return NetworkDeviceWireless(device_path, get_system_bus())
    return None


async def get_wifi_device_state() -> NetState:
    wifi_device = await get_wifi_device()
    if wifi_device is None:
        return NetState.UNKNOWN

    state = await wifi_device.state
    if state is DeviceState.UNKNOWN:
        return NetState.UNKNOWN
    if state in (
        DeviceState.DISCONNECTED,
        DeviceState.UNMANAGED,
        DeviceState.UNAVAILABLE,
        DeviceState.FAILED,
    ):
        return NetState.DISCONNECTED
    if state in (DeviceState.NEED_AUTH,):
        return NetState.NEEDS_ATTENTION
    if state in (
        DeviceState.DEACTIVATING,
        DeviceState.PREPARE,
        DeviceState.CONFIG,
        DeviceState.IP_CONFIG,
        DeviceState.IP_CHECK,
        DeviceState.SECONDARIES,
    ):
        return NetState.PENDING
    if state == DeviceState.ACTIVATED:
        return NetState.CONNECTED

    return NetState.UNKNOWN


@debounce(wait=0.5, options=DebounceOptions(trailing=True, time_window=2))
async def request_scan() -> None:
    wifi_device = await get_wifi_device()
    if wifi_device:
        await wait_for(wifi_device.request_scan({}))


async def get_access_points() -> list[AccessPoint]:
    wifi_device = await get_wifi_device()
    if wifi_device is None:
        return []

    access_points = await wait_for(
        wifi_device.access_points,
    )
    return [
        AccessPoint(access_point_path, get_system_bus())
        for access_point_path in access_points
    ]


async def get_active_access_point() -> AccessPoint | None:
    wifi_device = await get_wifi_device()
    if wifi_device is None:
        return None

    active_access_point = await wait_for(
        wifi_device.active_access_point,
    )
    if not active_access_point or active_access_point == '/':
        return None

    return AccessPoint(active_access_point, get_system_bus())


async def get_active_access_point_ssid() -> str | None:
    active_access_point = await get_active_access_point()
    if not active_access_point:
        return None

    return (
        await wait_for(
            active_access_point.ssid,
        )
    ).decode('utf-8')


async def get_active_connection() -> ActiveConnection | None:
    wifi_device = await get_wifi_device()
    if wifi_device is None:
        return None

    active_connection = await wait_for(
        wifi_device.active_connection,
    )
    if not active_connection or active_connection == '/':
        return None

    return ActiveConnection(active_connection, get_system_bus())


async def get_active_connection_state() -> ConnectionState:
    active_connection = await get_active_connection()
    if not active_connection:
        return ConnectionState.UNKNOWN

    active_connection_state = (
        cast('SdBusConnectionState', await active_connection.state)
        if active_connection
        else None
    )

    return {
        SdBusConnectionState.ACTIVATED: ConnectionState.CONNECTED,
        SdBusConnectionState.ACTIVATING: ConnectionState.CONNECTING,
        SdBusConnectionState.DEACTIVATED: ConnectionState.DISCONNECTED,
        SdBusConnectionState.DEACTIVATING: ConnectionState.DISCONNECTED,
        SdBusConnectionState.UNKNOWN: ConnectionState.UNKNOWN,
        None: ConnectionState.UNKNOWN,
    }[active_connection_state]


async def get_active_connection_ssid() -> str | None:
    active_connection = await get_active_connection()
    if not active_connection:
        return None

    try:
        connection = NetworkConnectionSettings(await active_connection.connection)

        settings = await connection.get_settings()
        return settings['802-11-wireless']['ssid'][1].decode('utf-8')
    except dbus_exceptions.DbusUnknownMethodError:
        report_service_error()
        return None


async def get_saved_ssids() -> list[str]:
    network_manager_settings = NetworkManagerSettings(get_system_bus())
    connections = [
        NetworkConnectionSettings(i, get_system_bus())
        for i in await wait_for(
            network_manager_settings.connections,
        )
    ]
    connections_settings = [
        await wait_for(
            i.get_settings(),
        )
        for i in connections
    ]
    return [
        settings['802-11-wireless']['ssid'][1].decode('utf-8')
        for settings in connections_settings
        if '802-11-wireless' in settings
    ]


async def wait_for_device_available(
    wifi_device: NetworkDeviceWireless,
) -> None:
    """Wait for WiFi device to become available.

    After enabling wireless, the device may be in UNAVAILABLE or UNMANAGED state.
    This function polls the device state until it becomes ready for connections.

    Args:
    ----
        wifi_device: The WiFi device to check

    Raises:
    ------
        WiFiDeviceNotAvailableError: If device does not become available after retries

    """
    from ubo_app.logger import logger

    async for attempt in AsyncRetrying(
        stop=stop_after_attempt(5),
        wait=wait_fixed(1),
        reraise=True,
    ):
        with attempt:
            state = await wifi_device.state
            if state in (
                DeviceState.UNAVAILABLE,
                DeviceState.UNMANAGED,
                DeviceState.UNKNOWN,
            ):
                logger.debug(
                    'Waiting for WiFi device to become available',
                    extra={
                        'state': state,
                        'attempt': attempt.retry_state.attempt_number,
                    },
                )
                raise WiFiDeviceNotAvailableError

            logger.debug(
                'WiFi device became available',
                extra={'state': state, 'attempt': attempt.retry_state.attempt_number},
            )


async def add_wireless_connection(
    ssid: str,
    password: str,
    type: WiFiType,
    *,
    hidden: bool | None = False,
) -> None:
    network_manager = NetworkManager(get_system_bus())
    await network_manager.wireless_enabled.set_async(True)

    wifi_device = await get_wifi_device()
    if not wifi_device:
        logger.error('WiFi device not found')
        return

    # Wait for device to become available after enabling wireless
    try:
        await wait_for_device_available(wifi_device)
    except WiFiDeviceNotAvailableError:
        logger.warning(
            'WiFi device did not become available, connection may fail',
        )
    access_points = [
        (
            access_point,
            await wait_for(
                access_point.ssid,
            ),
        )
        for access_point in await get_access_points()
    ]
    access_point = next(
        (
            access_point
            for access_point, ssid_ in access_points
            if ssid_.decode('utf8') == ssid
        ),
        None,
    )

    if type == WiFiType.NOPASS:
        security = {
            'key-mgmt': ('s', 'none'),
            'auth-alg': ('s', 'open'),
        }
    elif type == WiFiType.WEP:
        security = {
            'key-mgmt': ('s', 'none'),
            'auth-alg': ('s', 'open'),
            'psk': ('s', password),
        }
    elif type in (WiFiType.WPA, WiFiType.WPA2):
        security = {
            'key-mgmt': ('s', 'wpa-psk'),
            'auth-alg': ('s', 'open'),
            'psk': ('s', password),
        }

    properties: NetworkManagerConnectionProperties = {
        'connection': {
            'id': ('s', ssid),
            'uuid': ('s', str(uuid.uuid4())),
            'type': ('s', '802-11-wireless'),
            'autoconnect': ('b', True),
        },
        '802-11-wireless': {
            'mode': ('s', 'infrastructure'),
            'security': ('s', '802-11-wireless-security'),
            'ssid': ('ay', ssid.encode('utf-8')),
            'hidden': ('b', hidden),
        },
        '802-11-wireless-security': security,
        'ipv4': {'method': ('s', 'auto')},
        'ipv6': {'method': ('s', 'auto')},
    }

    connection = await wait_for(
        network_manager.add_and_activate_connection(
            properties,
            inspect_dbus_path(wifi_device),
            inspect_dbus_path(access_point) if access_point else '/',
        ),
    )

    logger.debug('Connection added', extra={'connection': connection})


async def connect_wireless_connection(ssid: str) -> None:
    network_manager = NetworkManager(get_system_bus())
    await network_manager.wireless_enabled.set_async(True)

    wifi_device = await get_wifi_device()

    if not wifi_device:
        return

    network_manager_settings = NetworkManagerSettings(get_system_bus())
    connections = [
        NetworkConnectionSettings(path, get_system_bus())
        for path in await wait_for(
            network_manager_settings.connections,
        )
    ]
    connections_settings = [
        (
            await wait_for(
                connection.get_settings(),
            ),
            inspect_dbus_path(connection),
        )
        for connection in connections
    ]
    desired_connection = next(
        (
            path
            for settings, path in connections_settings
            if '802-11-wireless' in settings
            and settings['802-11-wireless']['ssid'][1].decode('utf-8') == ssid
        ),
        None,
    )

    if not desired_connection:
        return

    await wait_for(
        network_manager.activate_connection(
            desired_connection,
            inspect_dbus_path(wifi_device),
        ),
    )


async def disconnect_wireless_connection() -> None:
    wifi_device = await get_wifi_device()
    if not wifi_device:
        return

    network_manager = NetworkManager(get_system_bus())
    await wait_for(
        network_manager.deactivate_connection(
            await wait_for(
                wifi_device.active_connection,
            ),
        ),
    )


async def forget_wireless_connection(ssid: str) -> None:
    network_manager_settings = NetworkManagerSettings(get_system_bus())

    for connection_path in await wait_for(
        network_manager_settings.connections,
    ):
        network_connection_settings = NetworkConnectionSettings(
            connection_path,
            get_system_bus(),
        )
        settings = await wait_for(
            network_connection_settings.get_settings(),
        )
        if (
            '802-11-wireless' in settings
            and settings['802-11-wireless']['ssid'][1].decode('utf-8') == ssid
        ):
            await wait_for(network_connection_settings.delete())
            store.dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        title=f'"{ssid}" Deleted',
                        content=f"""WiFi connection with ssid "{
                            ssid
                        }" was deleted successfully""",
                        display_type=NotificationDisplayType.FLASH,
                        color=DANGER_COLOR,
                        icon='󱛅',
                        chime=Chime.DONE,
                    ),
                ),
            )


async def get_connections() -> list[WiFiConnection]:
    # It is need as this action is not atomic and the active_connection may not be
    # available when active_connection.state is queried
    for _ in range(RETRIES):
        with contextlib.suppress(Exception):
            active_connection_state = await get_active_connection_state()
            active_connection_ssid = await get_active_connection_ssid()
            saved_ssids = await get_saved_ssids()
            access_point_by_ssids = {
                (
                    await wait_for(
                        i.ssid,
                    )
                ).decode('utf-8'): i
                for i in await get_access_points()
            }

            return [
                WiFiConnection(
                    ssid=ssid,
                    signal_strength=await wait_for(
                        access_point_by_ssids[ssid].strength,
                    )
                    if ssid in access_point_by_ssids
                    else 0,
                    state=active_connection_state
                    if active_connection_ssid == ssid
                    else ConnectionState.DISCONNECTED,
                )
                for ssid in saved_ssids
            ]
        await asyncio.sleep(0.5)
    return []



"""Shared utilities for Docker app operations (containers and compositions)."""

from __future__ import annotations

import json
from inspect import iscoroutine
from typing import TYPE_CHECKING

from docker_composition import COMPOSITIONS_PATH

from ubo_app.logger import logger
from ubo_app.store.main import store
from ubo_app.store.services.docker import DockerImageUpdateMetadataAction

if TYPE_CHECKING:
    from docker_images import ContainerEntry


async def prepare_app(container: ContainerEntry) -> bool:
    """Prepare an app (container or composition) for use.

    This function:
    1. Checks if a prepare function exists for the app
    2. Calls it (handling both sync and async functions)
    3. Checks the boolean result
    4. For compositions, updates metadata in Redux store
    5. Returns success/failure

    Parameters
    ----------
    container : ContainerEntry
        The container entry to prepare

    Returns
    -------
    bool
        True if preparation succeeded or no preparation needed, False otherwise

    """
    prepare_function = container.prepare
    if not prepare_function:
        return True

    logger.info(
        'Preparing app',
        extra={'image': container.id, 'is_composition': container.is_composition},
    )

    result = prepare_function()
    if iscoroutine(result):
        result = await result

    if not result:
        logger.error('Failed to prepare app', extra={'image': container.id})
        return False

    # Update metadata for compositions
    if container.is_composition:
        await update_composition_metadata(container)

    return True


async def update_composition_metadata(container: ContainerEntry) -> None:
    """Load and dispatch metadata for a composition.

    Reads the metadata.json file created by the prepare function
    and dispatches an action to update the Redux store with instructions
    and other metadata.

    Parameters
    ----------
    container : ContainerEntry
        The container entry for the composition

    """
    metadata_path = COMPOSITIONS_PATH / container.id / 'metadata.json'
    if not metadata_path.exists():
        logger.warning(
            'Metadata file not found for composition',
            extra={'image': container.id, 'path': str(metadata_path)},
        )
        return

    try:
        metadata = json.load(metadata_path.open())
        store.dispatch(
            DockerImageUpdateMetadataAction(
                image=container.id,
                instructions=metadata.get('instructions'),
            ),
        )
        logger.info(
            'Updated composition metadata',
            extra={
                'image': container.id,
                'has_instructions': bool(metadata.get('instructions')),
            },
        )
    except Exception:
        logger.exception(
            'Failed to load composition metadata',
            extra={'image': container.id, 'path': str(metadata_path)},
        )



"""Utility functions for dataclasses."""

from __future__ import annotations

import dataclasses
import sys
from typing import TYPE_CHECKING

from typing_extensions import TypeVar

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable

T = TypeVar('T', infer_variance=True)
_NOT_PROVIDED = object()


_FactoryType: type | None = None


def _extract_factory_type() -> None:
    """Extract the type of the factory function."""
    parent_frame = sys._getframe().f_back  # noqa: SLF001
    if not parent_frame:
        msg = 'No parent frame found'
        raise RuntimeError(msg)
    factory_instance = parent_frame.f_locals['sample_factory_field']
    global _FactoryType  # noqa: PLW0603
    _FactoryType = type(factory_instance)


@dataclasses.dataclass
class _Sample:
    sample_factory_field: str = dataclasses.field(default_factory=lambda: 'some_value')
    watcher_field: None = dataclasses.field(default_factory=_extract_factory_type)


_Sample()


def default_provider(
    required_fields: Iterable[str],
    provider: Callable[..., T],
) -> Callable[[], T]:
    """Provide value for a dataclass field based on a combination of other fields."""
    # WARNING: Dirty hack ahead
    # This is to set the default value of a dataclass field based on the
    # provided/default value of a set of other dataclass fields

    def default_provider() -> T:
        parent_frame = sys._getframe().f_back  # noqa: SLF001
        values = []
        for field in required_fields:
            if (
                not parent_frame
                or (value := parent_frame.f_locals.get(field, _NOT_PROVIDED))
                is _NOT_PROVIDED
            ):
                msg = f'No {field} provided'
                raise ValueError(msg)
            if _FactoryType and isinstance(value, _FactoryType):
                value = None
            values.append(value)

        return provider(*values)

    return default_provider



"""Transform a Python module into a proto file objects."""

from __future__ import annotations

import ast
import re
from typing import TYPE_CHECKING, Any

import betterproto.casing
from ubo_app.rpc.generator._types import (
    _BasicType,
    _DictType,
    _ListType,
    _OptionalType,
    _SetType,
    _Type,
    _UnionType,
    global_enums,
    global_messages,
    global_types,
)
from ubo_app.rpc.message_to_object import (
    META_FIELD_PREFIX_PACKAGE_NAME,
    META_FIELD_PREFIX_PACKAGE_NAME_INDEX,
)

if TYPE_CHECKING:
    from types import ModuleType


class _ProtoGenerator(ast.NodeVisitor):
    def __init__(
        self: _ProtoGenerator,
        module: ModuleType,
        actions: list,
        events: list,
    ) -> None:
        self.messages: dict[str, list[tuple[str, _Type]]] = {}
        self.enums: dict[str, list[tuple[str, Any]]] = {}
        self.types: dict[str, _Type] = {}
        self.module = module
        self.actions = actions
        self.events = events
        self.package_name = module.__name__

    def visit_ClassDef(self: _ProtoGenerator, node: ast.ClassDef) -> None:
        if any(
            base.id in ('Enum', 'StrEnum', 'IntEnum', 'Flag', 'IntFlag')
            for base in node.bases
            if isinstance(base, ast.Name)
        ):
            self.process_enum(node)
        else:
            self.process_class(node)
        self.generic_visit(node)

    def visit_AnnAssign(self: _ProtoGenerator, node: ast.AnnAssign) -> None:
        if (
            isinstance(node.target, ast.Name)
            and isinstance(node.annotation, ast.Name)
            and node.annotation.id == 'TypeAlias'
            and node.value
        ):
            field_name = node.target.id
            try:
                field_type = self.get_field_type(value=node.value)
            except TypeError as e:
                if 'Callable types are not supported' in str(e):
                    return
                raise
            self.types[field_name] = field_type
            if field_name in global_types:
                msg = (
                    f'Type "{field_name}" is already defined in '
                    f'"{global_types[field_name]}"'
                )
                raise SyntaxError(msg)
            global_types[field_name] = self.package_name
        self.generic_visit(node)

    def visit_Assign(self: _ProtoGenerator, node: ast.Assign) -> None:
        if (
            isinstance(node.value, ast.Call)
            and isinstance(node.value.func, ast.Name)
            and node.value.func.id == 'TypeVar'
            and isinstance(node.targets[0], ast.Name)
        ):
            field_name = node.targets[0].id
            bound = next(
                (
                    constraint
                    for constraint in node.value.keywords
                    if constraint.arg == 'bound'
                ),
                None,
            )
            if bound:
                bound_type = self.get_field_type(value=bound.value)
                if self.types.get(node.targets[0].id):
                    msg = f'''Type "{field_name}" is already defined as "{
                        self.types[node.targets[0].id]
                    }"'''
                    raise SyntaxError(msg)
                if not isinstance(bound_type, _UnionType):
                    bound_type = _UnionType.from_(types=[bound_type])
                self.types[field_name] = bound_type
                if field_name in global_types:
                    msg = (
                        f'Type "{field_name}" is already defined in '
                        f'"{global_types[field_name]}"'
                    )
                    raise SyntaxError(msg)
                global_types[field_name] = self.package_name
        self.generic_visit(node)

    def process_class(self: _ProtoGenerator, node: ast.ClassDef) -> None:  # noqa: C901
        message_name = node.name
        fields: list[tuple[str, _Type]] = []
        for n in node.body:
            if isinstance(n, ast.AnnAssign) and isinstance(n.target, ast.Name):
                field_name = n.target.id
                try:
                    field_type = self.get_field_type(value=n.annotation)
                    if n.value is not None and not isinstance(
                        field_type,
                        _OptionalType,
                    ):
                        field_type = _OptionalType(type=field_type)
                except TypeError as e:
                    if 'Callable types are not supported' in str(
                        e,
                    ) or 'ClassVar is not supported' in str(e):
                        continue
                    raise

                fields.append((field_name, field_type))
        if node.bases:
            for base in node.bases:
                if not isinstance(base, ast.Name):
                    continue
                base_name = base.id
                if base_name in global_messages:
                    fields.extend(
                        [
                            field
                            for field in global_messages[base_name][1]
                            if not any(field[0] == field_[0] for field_ in fields)
                        ],
                    )
        self.messages[message_name] = fields
        global_messages[message_name] = (self.package_name, fields)
        if message_name.endswith('Action'):
            self.actions.append((message_name, self.package_name))
        if message_name.endswith('Event'):
            self.events.append((message_name, self.package_name))

    def process_enum(self: _ProtoGenerator, node: ast.ClassDef) -> None:
        enum_name = node.name
        if enum_name in global_types:
            msg = (
                f'Enum "{enum_name}" is already defined in "{global_types[enum_name]}"'
            )
            raise SyntaxError(msg)
        values: list[tuple[str, Any]] = []
        for n in node.body:
            if isinstance(n, ast.Assign) and isinstance(n.targets[0], ast.Name):
                value_name = n.targets[0].id
                value = n.value.value if isinstance(n.value, ast.Constant) else None
                values.append((value_name, value))
        self.enums[enum_name] = values
        global_enums[enum_name] = self.package_name

    def get_field_type(  # noqa: C901, PLR0912, PLR0915
        self: _ProtoGenerator,
        *,
        value: ast.AST,
    ) -> _Type:
        if isinstance(value, ast.Name):
            if value.id == 'str':
                return _BasicType(type='string')
            if value.id == 'int':
                return _BasicType(type='int64')
            if value.id == 'float':
                return _BasicType(type='float')
            if value.id == 'bool':
                return _BasicType(type='bool')
            if value.id == 'bytes':
                return _BasicType(type='bytes')
            if value.id == 'datetime':
                return _BasicType(type='int64')
            if value.id == 'Path':
                return _BasicType(type='string')
            return _BasicType(type=value.id)
        if isinstance(value, ast.Constant) and value.value is None:
            return _BasicType(type='None')
        if isinstance(value, ast.Subscript):
            if isinstance(value.value, ast.Name):
                if value.value.id in ('Mapping', 'dict') and isinstance(
                    value.slice,
                    ast.Tuple,
                ):
                    key_type = self.get_field_type(value=value.slice.elts[0])
                    value_type = self.get_field_type(value=value.slice.elts[1])
                    return _DictType(key_type=key_type, value_type=value_type)

                if value.value.id in ('Sequence', 'list'):
                    return _ListType(type=self.get_field_type(value=value.slice))

                if value.value.id == 'set':
                    return _SetType(type=self.get_field_type(value=value.slice))

                if value.value.id == 'tuple' and isinstance(value.slice, ast.Tuple):
                    if (
                        len(value.slice.elts) == 2  # noqa: PLR2004
                        and isinstance(value.slice.elts[1], ast.Constant)
                        and value.slice.elts[1].value is ...
                    ):
                        types = {self.get_field_type(value=value.slice.elts[0])}
                    else:
                        types = {
                            self.get_field_type(value=elt) for elt in value.slice.elts
                        }

                    return _ListType(
                        type=_UnionType.from_(
                            types=list(types),
                        ),
                    )

                if value.value.id in self.types or value.value.id in self.messages:
                    msg = 'Generic types are not supported {value.value.id}'
                    raise TypeError(msg)

                if value.value.id == 'type':
                    return _BasicType(type='string')

                if value.value.id == 'Callable':
                    msg = 'Callable types are not supported'
                    raise TypeError(msg)

                if value.value.id == 'IO':
                    return _BasicType(type='bytes')

                if value.value.id == 'ClassVar':
                    msg = 'ClassVar is not supported'
                    raise TypeError(msg)

                msg = (
                    f'Unsupported subscript type: {value.value.id} '
                    f'- file: {self.module} - line: {value.lineno}'
                )
                raise TypeError(msg)

            msg = (
                f'Unsupported subscript type: {value.value} {value.slice} '
                f'- file: {self.module} - line: {value.lineno}'
            )
            raise TypeError(msg)

        if isinstance(value, ast.BinOp) and isinstance(value.op, ast.BitOr):
            types: set[_Type] = set()
            try:
                type = self.get_field_type(value=value.left)
                if isinstance(type, _UnionType):
                    types = types.union(type.types)
                else:
                    types.add(type)
            except TypeError as e:
                if 'Callable types are not supported' not in str(e):
                    raise
            try:
                type = self.get_field_type(value=value.right)
                if isinstance(type, _UnionType):
                    types = types.union(type.types)
                else:
                    types.add(type)
            except TypeError as e:
                if 'Callable types are not supported' not in str(e):
                    raise
            is_optional = any(
                isinstance(type, _BasicType) and type.type == 'None' for type in types
            )
            types_ = [
                type
                for type in types
                if not isinstance(type, _BasicType) or type.type != 'None'
            ]
            type = types_[0] if len(types_) == 1 else _UnionType.from_(types=types_)
            if is_optional:
                if isinstance(type, _OptionalType):
                    return type
                return _OptionalType(type=type)
            return type

        msg = f'Unsupported field type: {value}'
        raise TypeError(msg)

    def generate_proto(self: _ProtoGenerator) -> str:  # noqa: C901
        try:
            proto = ''
            for enum_name, values in self.enums.items():
                proto += f'enum {enum_name} {{\n'
                proto += f"""  {betterproto.casing.snake_case(enum_name).upper()}_{
                    self.package_name.replace('.', '_dot_').upper()
                }_UNSPECIFIED = 0;\n"""
                for i, (value_name, _) in enumerate(values, 0):
                    proto += f"""  {betterproto.casing.snake_case(enum_name).upper()}_{
                        value_name
                    } = {i + 1};\n"""
                proto += '}\n\n'
            for message_name, fields in self.messages.items():
                proto += f'message {message_name} {{\n'
                proto += f"""  option (package_info.v1.package_name) = "{
                    self.package_name
                }";\n"""
                proto += f'  optional string {META_FIELD_PREFIX_PACKAGE_NAME}'
                proto += f"""{self.package_name.replace('.', '_dot_')} = {
                    META_FIELD_PREFIX_PACKAGE_NAME_INDEX
                };\n"""
                for field_name, field_type in fields:
                    proto += re.sub(
                        r'\n(?=.)',
                        '\n',
                        field_type.get_definitions(
                            field_name,
                            current_package=self.package_name,
                        ),
                    )
                for i, (field_name, field_type) in enumerate(fields, 2):
                    try:
                        proto += f"""  {
                            field_type.get_proto(
                                field_name,
                                current_package=self.package_name,
                            )
                        } {field_name} = {i};\n"""
                    except TypeError as exception:
                        if 'Empty Union' in str(exception):
                            continue
                        if 'Unknown type' in str(exception):
                            continue
                        raise
                proto += '}\n\n'
            for name, field_type in self.types.items():
                proto += field_type.get_embedded_definitions(
                    name,
                    current_package=self.package_name,
                )
        except TypeError as e:
            msg = f'Error in {self.package_name}'
            raise TypeError(msg) from e
        else:
            return proto



"""Scheduler for Redux store."""

from __future__ import annotations

import asyncio
import contextlib
import threading
from typing import TYPE_CHECKING

from ubo_app.constants import DEBUG_SCHEDULER
from ubo_app.utils.error_handlers import loop_exception_handler

if TYPE_CHECKING:
    from collections.abc import Callable


class Scheduler(threading.Thread):
    """Store scheduler running in a separate thread."""

    # Freeze detection thresholds
    TICK_GAP_THRESHOLD = 1.0  # Log if gap between ticks exceeds 1 second
    CALLBACK_DURATION_THRESHOLD = 0.5  # Log if callback takes > 500ms
    HEARTBEAT_INTERVAL = 30.0  # Periodic health log every 30 seconds

    def __init__(self: Scheduler) -> None:
        """Initialize the scheduler."""
        super().__init__()
        self.name = 'Scheduler Thread'
        self.stopped = False
        self._callbacks: list[tuple[Callable[[], None], float]] = []
        self.loop = asyncio.new_event_loop()
        self.loop.set_exception_handler(loop_exception_handler)
        self.tasks: set[asyncio.Task] = set()
        # Freeze detection state (only used when DEBUG_SCHEDULER is enabled)
        if DEBUG_SCHEDULER:
            self._last_tick_time: float = 0.0
            self._last_heartbeat_time: float = 0.0
            self._tick_count: int = 0
            self._freeze_count: int = 0
            self._start_time: float = 0.0  # Set when first tick runs

    def run(self: Scheduler) -> None:
        """Run the scheduler."""
        from ubo_app.logger import logger

        logger.info('Starting scheduler async loop')
        try:
            self.loop.run_forever()
        except:  # noqa: E722
            logger.info('Scheduler ran into an error and stopped', exc_info=True)
        else:
            logger.info('Scheduler stopped gracefully')

    def set(
        self: Scheduler,
        callback: Callable[[], None],
        *,
        interval: bool,
        delay_duration: float | None = None,
    ) -> None:
        """Start the scheduler."""
        self.loop.call_soon_threadsafe(
            self.loop.create_task,
            self.call_callback(
                callback,
                interval=interval,
                delay_duration=delay_duration,
            ),
        )

    async def call_callback(
        self: Scheduler,
        callback: Callable[[], None],
        *,
        interval: bool,
        delay_duration: float | None,
        last_call: float | None = None,
    ) -> None:
        """Call the callback function."""
        if self.stopped:
            return
        if delay_duration is None:
            delay_duration = 0.025
        now = self.loop.time()
        if last_call is None:
            last_call = now
        required_sleep = max(
            delay_duration - max(now - last_call - delay_duration, 0),
            0,
        )
        await asyncio.sleep(required_sleep)

        if DEBUG_SCHEDULER:
            await self._call_callback_with_monitoring(callback)
        else:
            try:
                callback()
            except Exception:
                from ubo_app.logger import logger

                logger.exception('Error in store heartbeat callback')

        if interval:
            self.tasks.add(
                self.loop.create_task(
                    self.call_callback(
                        callback,
                        interval=interval,
                        delay_duration=delay_duration,
                        last_call=now,
                    ),
                ),
            )
            self.tasks = {task for task in self.tasks if not task.done()}

    async def _call_callback_with_monitoring(
        self: Scheduler,
        callback: Callable[[], None],
    ) -> None:
        """Call callback with freeze detection and monitoring (DEBUG_SCHEDULER only)."""
        import time

        # Freeze detection: check gap since last tick
        current_time = time.time()
        tick_gap = (
            current_time - self._last_tick_time if self._last_tick_time > 0 else 0
        )

        try:
            start_time = time.time()
            callback()
            callback_duration = time.time() - start_time

            self._tick_count += 1
            self._last_tick_time = time.time()

            # Track when scheduler started (first tick)
            if self._start_time == 0:
                self._start_time = current_time

            # Log only when freeze detected (tick gap > threshold)
            if tick_gap > self.TICK_GAP_THRESHOLD:
                from ubo_app.logger import logger

                self._freeze_count += 1
                logger.warning(
                    'Scheduler freeze detected',
                    extra={
                        'tick_gap_seconds': round(tick_gap, 2),
                        'callback_duration_ms': round(callback_duration * 1000, 1),
                        'freeze_count': self._freeze_count,
                        'tick_count': self._tick_count,
                    },
                )

            # Log if callback itself took too long
            elif callback_duration > self.CALLBACK_DURATION_THRESHOLD:
                from ubo_app.logger import logger

                logger.warning(
                    'Scheduler callback slow',
                    extra={
                        'callback_duration_ms': round(callback_duration * 1000, 1),
                        'tick_count': self._tick_count,
                    },
                )

            # Periodic heartbeat (every 30s) - confirms scheduler is alive
            elif current_time - self._last_heartbeat_time > self.HEARTBEAT_INTERVAL:
                from ubo_app.logger import logger

                self._last_heartbeat_time = current_time
                logger.debug(
                    'Scheduler heartbeat',
                    extra={
                        'tick_count': self._tick_count,
                        'freeze_count': self._freeze_count,
                    },
                )

        except Exception:
            from ubo_app.logger import logger

            logger.exception('Error in store heartbeat callback')

    async def shutdown(self: Scheduler) -> None:
        """Stop the scheduler gracefully."""
        self.stopped = True

        # Log summary on shutdown (only when DEBUG_SCHEDULER is enabled)
        if DEBUG_SCHEDULER and self._start_time > 0:
            import time

            from ubo_app.logger import logger

            runtime = time.time() - self._start_time
            expected_ticks = int(runtime / 0.025)
            efficiency = (
                round(self._tick_count / expected_ticks * 100, 1)
                if expected_ticks > 0
                else 0
            )
            logger.debug(
                'Scheduler shutdown summary',
                extra={
                    'runtime_seconds': round(runtime, 1),
                    'tick_count': self._tick_count,
                    'expected_ticks': expected_ticks,
                    'efficiency_percent': efficiency,
                    'freeze_count': self._freeze_count,
                },
            )

        await asyncio.sleep(0.05)
        with contextlib.suppress(BaseException):
            await asyncio.wait_for(asyncio.gather(*self.tasks), timeout=0.1)
        self.tasks.clear()
        self.loop.stop()

    def stop(self: Scheduler) -> None:
        """Schedule the shutdown of the scheduler."""
        self.loop.call_soon_threadsafe(self.loop.create_task, self.shutdown())



"""Utility functions to work with the persistent storage."""

from __future__ import annotations

import json
import json.decoder
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar, cast, overload

import fasteners

from ubo_app.constants import PERSISTENT_STORE_PATH

if TYPE_CHECKING:
    from collections.abc import Callable

    from ubo_app.store.main import RootState

T = TypeVar('T')

persistent_store_lock = fasteners.ReaderWriterLock()


def register_persistent_store(
    key: str,
    selector: Callable[[RootState], T],
) -> None:
    """Register a part of the store to be persistent in the file system."""
    from ubo_app.store.main import store

    @store.autorun(selector)
    async def _(value: T) -> None:
        if value is None:
            return
        with persistent_store_lock.write_lock():
            try:
                current_state = json.loads(Path(PERSISTENT_STORE_PATH).read_text())
            except (FileNotFoundError, json.decoder.JSONDecodeError):
                current_state = {}
            serialized_value = store.serialize_value(value)
            current_state[key] = serialized_value
            Path(PERSISTENT_STORE_PATH).write_text(json.dumps(current_state, indent=2))


@overload
def read_from_persistent_store(key: str) -> str: ...
@overload
def read_from_persistent_store(
    key: str,
    *,
    mapper: Callable[[str], T],
) -> T: ...
@overload
def read_from_persistent_store(
    key: str,
    *,
    output_type: type[T],
) -> T: ...
@overload
def read_from_persistent_store(
    key: str,
    *,
    default: T,
) -> T: ...
@overload
def read_from_persistent_store(
    key: str,
    *,
    default: T,
    mapper: Callable[[str], T],
) -> T: ...
@overload
def read_from_persistent_store(
    key: str,
    *,
    default: T,
    output_type: type[T],
) -> T: ...


def read_from_persistent_store(
    key: str,
    *,
    default: T | None = None,
    output_type: type[T] | None = None,
    mapper: Callable[[str], T] | None = None,
) -> T | None:
    """Read a part of the store from the file system."""
    from ubo_app.store.main import store

    if output_type is not None and mapper:
        msg = 'You cannot specify both `output_type` and `mapper` arguments.'
        raise ValueError(msg)

    try:
        with persistent_store_lock.read_lock():
            file_content = Path(PERSISTENT_STORE_PATH).read_text()
        current_state = json.loads(file_content)
    except (FileNotFoundError, json.decoder.JSONDecodeError):
        current_state = {}
    value = current_state.get(key)
    if value is None:
        return (
            (None if output_type is None else output_type())
            if default is None
            else default
        )

    if mapper:
        return mapper(value)
    return store.load_object(
        value,
        object_type=cast('type[T]', output_type),
    )
